# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import mock

from pypowervm import const as c
from pypowervm import exceptions as e
from pypowervm.tasks import vfc_mapper
from pypowervm.tests.tasks import util as tju
from pypowervm.tests.test_utils import test_wrapper_abc as twrap
from pypowervm.wrappers import virtual_io_server as pvm_vios

VIOS_FILE = 'fake_vios.txt'
VIOS_FEED = 'fake_vios_feed.txt'

FAKE_UUID = '42DF39A2-3A4A-4748-998F-25B15352E8A7'


class TestVFCMapper(unittest.TestCase):

    @mock.patch('pypowervm.wrappers.job.Job')
    def test_build_wwpn_pair(self, mock_job):
        mock_adpt = mock.MagicMock()
        mock_adpt.read.return_value = mock.Mock()

        # Mock out the job response
        job_w = mock.MagicMock()
        mock_job.wrap.return_value = job_w
        job_w.get_job_results_as_dict.return_value = {'wwpnList':
                                                      'a,b,c,d,e,f,g,h'}

        # Invoke and validate
        resp = vfc_mapper.build_wwpn_pair(mock_adpt, 'host_uuid', pair_count=4)
        self.assertEqual(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'], resp)

        # Make sure that the job was built properly
        mock_adpt.read.assert_called_once_with(
            'ManagedSystem', root_id='host_uuid', suffix_type=c.SUFFIX_TYPE_DO,
            suffix_parm=vfc_mapper._GET_NEXT_WWPNS)
        job_w.create_job_parameter.assert_called_once_with(
            'numberPairsRequested', '4')

    def test_find_vios_for_wwpn(self):
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FILE).entry)
        vios_feed_w = [vios_w]

        # Basic test
        vio_resp, p_resp = vfc_mapper.find_vios_for_wwpn(
            vios_feed_w, '10000090FA45473B')
        self.assertEqual(vios_w, vio_resp)
        self.assertIsNotNone(p_resp)

        # Validates the sanitized input
        vio_resp, p_resp = vfc_mapper.find_vios_for_wwpn(
            vios_feed_w, '10:00:00:90:fa:45:47:3b')
        self.assertEqual(vios_w, vio_resp)
        self.assertIsNotNone(p_resp)

        # Make sure a bad WWPN returns no result
        vio_resp, p_resp = vfc_mapper.find_vios_for_wwpn(
            vios_feed_w, '10:00:00:90:fa:45:47:3f')
        self.assertIsNone(vio_resp)
        self.assertIsNone(p_resp)

    def test_intersect_wwpns(self):
        list1 = ['AA:BB:CC:DD:EE:FF']
        list2 = {'aabbccddeeff', '1234567890'}
        self.assertEqual(list1, vfc_mapper.intersect_wwpns(list1, list2))

        # Full match
        list1 = {'aabbccddeeff', '1234567890'}
        list2 = ['AA:BB:CC:DD:EE:FF', '12:34:56:78:90']
        self.assertEqual(list1, set(vfc_mapper.intersect_wwpns(list1, list2)))

        # Second set as the limiter
        list1 = ['AA:BB:CC:DD:EE:FF', '12:34:56:78:90']
        list2 = {'aabbccddeeff'}
        self.assertEqual(['AA:BB:CC:DD:EE:FF'],
                         vfc_mapper.intersect_wwpns(list1, list2))

    def test_derive_npiv_map(self):
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FILE).entry)
        vios_wraps = [vios_w]

        # Subset the WWPNs on that VIOS
        p_wwpns = ['10000090FA45473B', '10:00:00:90:fa:45:17:58']

        # Virtual WWPNs can be faked, and simplified.
        v_port_wwpns = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

        # Run the derivation now
        resp = vfc_mapper.derive_npiv_map(vios_wraps, p_wwpns, v_port_wwpns)
        self.assertIsNotNone(resp)
        self.assertEqual(5, len(resp))

        # Make sure we only get two unique keys back.
        unique_keys = set([i[0] for i in resp])
        self.assertEqual({'10000090FA45473B', '10000090FA451758'}, unique_keys)

    def test_derive_npiv_map_existing_preserve(self):
        # Use sample vios data with mappings.
        vios_file = 'fake_vios_mappings.txt'
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(vios_file).entry)
        vios_wraps = [vios_w]

        # Subset the WWPNs on that VIOS
        p_wwpns = ['10000090FA1B6898', '10000090FA1B6899']
        v_port_wwpns = ['c05076065a7c02e4', 'c05076065a7c02e5']
        candidates = vfc_mapper._find_ports_on_vio(vios_w, p_wwpns)
        for p_port in candidates:
            if p_port.wwpn == p_wwpns[1]:
                # Artificially inflate the free ports so that it would get
                # chosen for a newly created mapping, but first in list
                # would show up for a preserved mapping.
                p_port.set_parm_value('AvailablePorts', '64')

        # Run the derivation now
        resp = vfc_mapper.derive_npiv_map(vios_wraps, p_wwpns, v_port_wwpns,
                                          preserve=True)
        self.assertIsNotNone(resp)
        self.assertEqual(1, len(resp))

        # Make sure we only got the one phys port key back that has the
        # existing mapping.
        unique_keys = set([i[0] for i in resp])
        self.assertEqual({'10000090FA1B6898'}, unique_keys)

    def test_derive_npiv_map_existing_no_preserve(self):
        # Use sample vios data with mappings.
        vios_file = 'fake_vios_mappings.txt'
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(vios_file).entry)
        vios_wraps = [vios_w]

        # Subset the WWPNs on that VIOS
        p_wwpns = ['10000090FA1B6898', '10000090FA1B6899']
        v_port_wwpns = ['c05076065a7c02e4', 'c05076065a7c02e5']
        candidates = vfc_mapper._find_ports_on_vio(vios_w, p_wwpns)
        for p_port in candidates:
            if p_port.wwpn == p_wwpns[1]:
                # Artificially inflate the free ports so that it would get
                # chosen for a newly created mapping.
                p_port.set_parm_value('AvailablePorts', '64')

        # Run the derivation now
        resp = vfc_mapper.derive_npiv_map(vios_wraps, p_wwpns, v_port_wwpns,
                                          preserve=False)
        self.assertIsNotNone(resp)
        self.assertEqual(1, len(resp))

        # Make sure we only got one phys port key back and it should *not*
        # match the existing mapping of 'preserve' testcase.
        unique_keys = set([i[0] for i in resp])
        self.assertEqual({'10000090FA1B6899'}, unique_keys)

    def test_derive_base_npiv_map(self):
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FILE).entry)
        vios_wraps = [vios_w]

        # Subset the WWPNs on that VIOS
        p_wwpns = ['10000090FA45473B', '10:00:00:90:fa:45:17:58']

        # Run the derivation now
        resp = vfc_mapper.derive_base_npiv_map(vios_wraps, p_wwpns, 5)
        self.assertIsNotNone(resp)
        self.assertEqual(5, len(resp))

        # Make sure we only get two unique keys back.
        unique_keys = set([i[0] for i in resp])
        self.assertEqual({'10000090FA45473B', '10000090FA451758'}, unique_keys)

        # Make sure we get the 'marker' back for the values.  Should now be
        # fused.
        values = set(i[1] for i in resp)
        self.assertEqual({vfc_mapper._FUSED_ANY_WWPN}, values)

    def test_derive_npiv_map_multi_vio(self):
        vios_wraps = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FEED))

        # Subset the WWPNs on that VIOS
        p_wwpns = ['10000090FA5371F2', '10000090FA53720A']

        # Virtual WWPNs can be faked, and simplified.
        v_port_wwpns = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

        # Run the derivation now
        resp = vfc_mapper.derive_npiv_map(vios_wraps, p_wwpns, v_port_wwpns)
        self.assertIsNotNone(resp)
        self.assertEqual(5, len(resp))

        # Make sure we only get two unique keys back.
        unique_keys = set([i[0] for i in resp])
        self.assertEqual(set(p_wwpns), unique_keys)

    def test_derive_npiv_map_failure(self):
        """Make sure we get a failure in the event of no candidates."""
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FILE).entry)
        vios_wraps = [vios_w]

        # Subset the WWPNs on that VIOS.  These WWPNs don't actually exist,
        # so the VIOSes passed in won't have these as candidate ports.
        p_wwpns = ['10000090FA45473bA', '10:00:00:90:fa:45:17:58A']

        # Virtual WWPNs can be faked, and simplified.
        v_port_wwpns = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

        # Run the derivation now
        self.assertRaises(e.UnableToFindFCPortMap, vfc_mapper.derive_npiv_map,
                          vios_wraps, p_wwpns, v_port_wwpns)

    def test_find_map_port(self):
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FILE).entry)

        # Happy path, should find the first port on the VIOS
        p1 = vfc_mapper._find_map_port(vios_w.pfc_ports, [])
        self.assertIsNotNone(p1)

        # Lets add a mapping where P1 is used.  Should not get that result
        # back.
        p2 = vfc_mapper._find_map_port(vios_w.pfc_ports, [(p1.wwpn, '')])
        self.assertIsNotNone(p2)
        self.assertNotEqual(p1, p2)

        # Now add a third and fourth port.  Same assertions.
        p3 = vfc_mapper._find_map_port(vios_w.pfc_ports, [(p1.wwpn, ''),
                                                          (p2.wwpn, '')])
        self.assertIsNotNone(p3)
        self.assertNotIn(p3, [p1, p2])

        p4 = vfc_mapper._find_map_port(vios_w.pfc_ports, [(p1.wwpn, ''),
                                                          (p2.wwpn, ''),
                                                          (p3.wwpn, '')])
        self.assertIsNotNone(p4)
        self.assertNotIn(p4, [p1, p2, p3])

        # Artificially inflate the use of other ports.
        port_use = [(p1.wwpn, ''), (p2.wwpn, ''), (p3.wwpn, ''), (p4.wwpn, ''),
                    (p1.wwpn, ''), (p2.wwpn, ''), (p4.wwpn, '')]
        p_temp = vfc_mapper._find_map_port(vios_w.pfc_ports, port_use)
        self.assertIsNotNone(p_temp)
        self.assertNotIn(p_temp, [p1, p2, p4])

    def test_fuse_vfc_ports(self):
        self.assertEqual(['A B'], vfc_mapper._fuse_vfc_ports(['a', 'b']))
        self.assertEqual(['AA BB'], vfc_mapper._fuse_vfc_ports(['a:a', 'b:b']))
        self.assertEqual(['A B', 'C D'],
                         vfc_mapper._fuse_vfc_ports(['a', 'b', 'c', 'd']))

    @mock.patch('pypowervm.tasks.vfc_mapper.derive_base_npiv_map')
    def test_build_migration_mappings_for_fabric(self, mock_derive):
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FILE).entry)
        vios_wraps = [vios_w]

        # Subset the WWPNs on that VIOS
        p_wwpns = ['10000090FA45473B', '10:00:00:90:fa:45:17:58']
        client_slots = ['1', '2']

        # The derive is non-deterministic.  That makes testing odd.  Force
        # a deterministic result.
        mock_derive.return_value = [('10000090FA451758', 'A A'),
                                    ('10000090FA45473B', 'B B')]

        # Build migration mappings success case
        resp = vfc_mapper.build_migration_mappings_for_fabric(
            vios_wraps, p_wwpns, client_slots)
        self.assertEqual(2, len(resp))
        self.assertEqual({'1/IO Server/1//fcs2', '2/IO Server/1//fcs1'},
                         set(resp))

    def test_build_migration_mappings_for_fabric_invalid_physical_port(self):
        vios_w = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FILE).entry)
        vios_wraps = [vios_w]

        # Invalid WWPNs should raise an error.
        p_wwpns = ['10000090FA45477B']
        client_slots = ['1', '2']

        # Build migration mappings success case
        self.assertRaises(e.UnableToFindFCPortMap,
                          vfc_mapper.build_migration_mappings_for_fabric,
                          vios_wraps, p_wwpns, client_slots)

    def test_build_migration_mappings(self):
        vios_wraps = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FEED))
        fabric_data = {'A': {'slots': [3, 4],
                             'p_port_wwpns': ["10000090FA5371F1",
                                              "10000090FA53720A"]},
                       'B': {'slots': [5, 6],
                             'p_port_wwpns': ["10000090FA5371F2",
                                              "10000090FA537209"]}}
        slot_peers = [[3, 5], [4, 6]]
        resp = vfc_mapper.build_migration_mappings(vios_wraps, fabric_data,
                                                   slot_peers)
        self.assertEqual(4, len(resp))
        self.assertEqual(set(resp), {'4/nimbus-ch03-p2-vios1/1//fcs1',
                                     '6/nimbus-ch03-p2-vios1/1//fcs0',
                                     '3/nimbus-ch03-p2-vios2/2//fcs0',
                                     '5/nimbus-ch03-p2-vios2/2//fcs1'})

        fabric_data = {'A': {'slots': [3],
                             'p_port_wwpns': ["10000090FA5371F1"]},
                       'B': {'slots': [5, 6],
                             'p_port_wwpns': ["10000090FA5371F2",
                                              "10000090FA537209"]}}
        slot_peers = [[3, 5], [6]]
        resp = vfc_mapper.build_migration_mappings(vios_wraps, fabric_data,
                                                   slot_peers)
        self.assertEqual(3, len(resp))
        self.assertEqual(set(resp), {'5/nimbus-ch03-p2-vios2/2//fcs1',
                                     '3/nimbus-ch03-p2-vios2/2//fcs0',
                                     '6/nimbus-ch03-p2-vios1/1//fcs0'})
        # Use invalid ports
        fabric_data = {'A': {'slots': [3],
                             'p_port_wwpns': ["10000090FA5371F1"]},
                       'B': {'slots': [5],
                             'p_port_wwpns': ["10000090FA537209"]}}
        slot_peers = [[3, 5]]
        self.assertRaises(e.UnableToFindFCPortMap,
                          vfc_mapper.build_migration_mappings,
                          vios_wraps, fabric_data, slot_peers)


class TestPortMappings(twrap.TestWrapper):
    file = VIOS_FEED
    wrapper_class_to_test = pvm_vios.VIOS
    mock_adapter_fx_args = {}

    def setUp(self):
        super(TestPortMappings, self).setUp()
        href_p = mock.patch('pypowervm.wrappers.virtual_io_server.VFCMapping.'
                            '_client_lpar_href')
        href = href_p.start()
        self.addCleanup(href_p.stop)
        href.return_value = 'fake_href'
        self.adpt.read.return_value = self.resp

    def test_find_vios_for_port_map(self):
        """Tests the find_vios_for_port_map method."""
        # Try off of the client WWPNs
        e0 = ('bad', 'c05076079cff08da c05076079cff08db')
        self.assertEqual(self.entries[0],
                         vfc_mapper.find_vios_for_port_map(self.entries, e0))

        # This WWPN is on the first VIOS
        e1 = ('10000090FA5371f1', 'a b')
        self.assertEqual(self.entries[0],
                         vfc_mapper.find_vios_for_port_map(self.entries, e1))

        # This WWPN is on the second VIOS
        e2 = ('10000090FA537209', 'a b')
        self.assertEqual(self.entries[1],
                         vfc_mapper.find_vios_for_port_map(self.entries, e2))

        # Try with a bad WWPN
        e3 = ('BAD', 'a b')
        self.assertIsNone(vfc_mapper.find_vios_for_port_map(self.entries, e3))

    def test_find_vios_for_vfc_wwpns(self):
        """Tests the find_vios_for_vfc_wwpns method."""
        # This WWPN is on the first VIOS
        v_wwpns = ['c05076079cff0e56', 'c05076079cff0e57']
        vios, vmap = vfc_mapper.find_vios_for_vfc_wwpns(self.entries, v_wwpns)
        self.assertEqual(self.entries[0], vios)
        self.assertEqual('10000090FA5371F2', vmap.backing_port.wwpn)

        # Have one of the ports be wrong
        v_wwpns = ['c05076079cff0e56', 'c05076079cff0e59']
        vios, vmap = vfc_mapper.find_vios_for_vfc_wwpns(self.entries, v_wwpns)
        self.assertIsNone(vios)
        self.assertIsNone(vmap)

        # Try odd formatting
        v_wwpns = ['C05076079cff0E56', 'c0:50:76:07:9c:ff:0E:57']
        vios, vmap = vfc_mapper.find_vios_for_vfc_wwpns(self.entries, v_wwpns)
        self.assertEqual(self.entries[0], vios)
        self.assertEqual('10000090FA5371F2', vmap.backing_port.wwpn)

        # Second VIOS
        v_wwpns = ['c05076079cff07ba', 'c05076079cff07bb']
        vios, vmap = vfc_mapper.find_vios_for_vfc_wwpns(self.entries, v_wwpns)
        self.assertEqual(self.entries[1], vios)
        self.assertEqual('10000090FA53720A', vmap.backing_port.wwpn)

        # Reverse WWPNs
        v_wwpns = ['c05076079cff07bb', 'c05076079cff07ba']
        vios, vmap = vfc_mapper.find_vios_for_vfc_wwpns(self.entries, v_wwpns)
        self.assertEqual(self.entries[1], vios)
        self.assertEqual('10000090FA53720A', vmap.backing_port.wwpn)

        # Set Type
        v_wwpns = {'c05076079cff07bb', 'c05076079cff07ba'}
        vios, vmap = vfc_mapper.find_vios_for_vfc_wwpns(self.entries, v_wwpns)
        self.assertEqual(self.entries[1], vios)
        self.assertEqual('10000090FA53720A', vmap.backing_port.wwpn)

    def test_find_pfc_wwpn_by_name(self):
        vio_w = self.entries[0]
        self.assertEqual('10000090FA5371F1',
                         vfc_mapper._find_pfc_wwpn_by_name(vio_w, 'fcs0'))
        self.assertIsNone(vfc_mapper._find_pfc_wwpn_by_name(vio_w, 'fcsX'))

    @mock.patch('lxml.etree.tostring')
    def test_add_port_bad_pfc(self, mock_tostring):
        """Validates that an error will be thrown with a bad pfc port."""
        # Build the mappings - the provided WWPN is bad
        vfc_map = ('10000090FA5371F9', '0 1')

        # Now call the add action.  This should log a warning.
        with self.assertLogs(vfc_mapper.__name__, level='WARNING'):
            self.assertRaises(e.UnableToDerivePhysicalPortForNPIV,
                              vfc_mapper.add_map,
                              self.entries[0], 'host_uuid', FAKE_UUID, vfc_map)
        mock_tostring.assert_called_once_with(
            self.entries[0].entry.element.element, pretty_print=True)

    def ensure_does_not_have_wwpns(self, vios_w, wwpns):
        for vfc_map in vios_w.vfc_mappings:
            if vfc_map.client_adapter is None:
                continue
            for c_wwpn in vfc_map.client_adapter.wwpns:
                if c_wwpn in wwpns:
                    self.fail("WWPN %s in client adapter" % vfc_mapper)

    def ensure_has_wwpns(self, vios_w, wwpns):
        for my_wwpn in wwpns:
            has_wwpn = False
            for vfc_map in vios_w.vfc_mappings:
                if vfc_map.client_adapter is None:
                    continue
                for c_wwpn in vfc_map.client_adapter.wwpns:
                    if c_wwpn == my_wwpn:
                        has_wwpn = True
                        break
            if not has_wwpn:
                self.fail("Unable to find WWPN %s" % my_wwpn)

    def test_find_maps(self):
        vwrap = self.entries[0]
        matches = vfc_mapper.find_maps(vwrap.vfc_mappings, 10)
        # Make sure we got the right ones
        self.assertEqual(
            ['U7895.43X.21EF9FB-V63-C3', 'U7895.43X.21EF9FB-V66-C4',
             'U7895.43X.21EF9FB-V62-C4', 'U7895.43X.21EF9FB-V10-C4'],
            [match.client_adapter.loc_code for match in matches])
        # Bogus LPAR ID
        self.assertEqual([], vfc_mapper.find_maps(vwrap.vfc_mappings, 1000))

        # Now try with UUID
        matches = vfc_mapper.find_maps(vwrap.vfc_mappings,
                                       '3ADDED46-B3A9-4E12-B6EC-8223421AF49B')
        self.assertEqual(
            ['U7895.43X.21EF9FB-V63-C3', 'U7895.43X.21EF9FB-V66-C4',
             'U7895.43X.21EF9FB-V62-C4', 'U7895.43X.21EF9FB-V10-C4'],
            [match.client_adapter.loc_code for match in matches])
        # Bogus LPAR UUID
        self.assertEqual([], vfc_mapper.find_maps(
            vwrap.vfc_mappings, '4BEEFD00-B3A9-4E12-B6EC-8223421AF49B'))

    def test_remove_maps(self):
        v_wrap = self.entries[0]
        len_before = len(v_wrap.vfc_mappings)
        resp_list = vfc_mapper.remove_maps(v_wrap, 10)
        expected_removals = {
            'U7895.43X.21EF9FB-V63-C3', 'U7895.43X.21EF9FB-V66-C4',
            'U7895.43X.21EF9FB-V62-C4', 'U7895.43X.21EF9FB-V10-C4'}
        self.assertEqual(
            set([el.client_adapter.loc_code for el in resp_list]),
            expected_removals)
        self.assertEqual(len_before - 4, len(v_wrap.vfc_mappings))

        # Make sure the remaining adapters do not have the remove codes.
        for remaining_map in v_wrap.vfc_mappings:
            if remaining_map.client_adapter is not None:
                self.assertNotIn(remaining_map.client_adapter.loc_code,
                                 expected_removals)

    def test_remove_maps_client_adpt(self):
        """Tests the remove_maps method, with the client_adpt input."""
        v_wrap = self.entries[0]
        len_before = len(v_wrap.vfc_mappings)

        c_adpt = vfc_mapper.find_maps(
            v_wrap.vfc_mappings, 10)[0].client_adapter

        resp_list = vfc_mapper.remove_maps(v_wrap, 10, client_adpt=c_adpt)
        expected_removals = {'U7895.43X.21EF9FB-V63-C3'}
        self.assertEqual(
            set([el.client_adapter.loc_code for el in resp_list]),
            expected_removals)
        self.assertEqual(len_before - 1, len(v_wrap.vfc_mappings))

        # Make sure the remaining adapters do not have the remove codes.
        for remaining_map in v_wrap.vfc_mappings:
            if remaining_map.client_adapter is not None:
                self.assertNotIn(remaining_map.client_adapter.loc_code,
                                 expected_removals)

    def test_has_client_wwpns(self):
        v_wrap_1 = self.entries[0]
        v_wrap_2 = self.entries[1]
        vio_w, vfc_map = vfc_mapper.has_client_wwpns(
            self.entries, ['c05076079cff0e56', 'c05076079cff0e57'])
        self.assertEqual(v_wrap_1, vio_w)
        self.assertEqual('10000090FA5371F2', vfc_map.backing_port.wwpn)

        # Second vios.  Reversed WWPNs.  Mixed Case.
        vio_w, vfc_map = vfc_mapper.has_client_wwpns(
            self.entries, ['c05076079cff0e83', 'c05076079cff0E82'])
        self.assertEqual(v_wrap_2, vio_w)
        self.assertEqual('10000090FA537209', vfc_map.backing_port.wwpn)

        # Not found.
        vio_w, vfc_map = vfc_mapper.has_client_wwpns(
            self.entries, ['AAA', 'bbb'])
        self.assertIsNone(vio_w)
        self.assertIsNone(vfc_map)


class TestAddRemoveMap(twrap.TestWrapper):
    file = VIOS_FEED
    wrapper_class_to_test = pvm_vios.VIOS
    mock_adapter_fx_args = {}

    def setUp(self):
        super(TestAddRemoveMap, self).setUp()
        href_p = mock.patch('pypowervm.wrappers.virtual_io_server.VFCMapping.'
                            'crt_related_href')
        href = href_p.start()
        self.addCleanup(href_p.stop)
        href.return_value = (
            'https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
            'e7344c5b-79b5-3e73-8f64-94821424bc25/LogicalPartition/'
            '3ADDED46-B3A9-4E12-B6EC-8223421AF49B')
        self.adpt.read.return_value = self.resp

        self.lpar_uuid = '3ADDED46-B3A9-4E12-B6EC-8223421AF49B'

    def test_add_remove_map_any_wwpn(self):
        """Tests a loop of add map/remove map when using _ANY_WWPN."""
        v_wrap = self.entries[0]
        len_before = len(v_wrap.vfc_mappings)

        # A fake mapping to the first IO Server
        p_map_vio1 = ('10000090FA5371F2', vfc_mapper._FUSED_ANY_WWPN)
        vfc_mapper.add_map(v_wrap, 'host_uuid', self.lpar_uuid, p_map_vio1)
        self.assertEqual(len_before + 1, len(v_wrap.vfc_mappings))

        # See if we can find that mapping.
        maps = vfc_mapper.find_maps(v_wrap.vfc_mappings, self.lpar_uuid,
                                    port_map=p_map_vio1)
        self.assertEqual(1, len(maps))

        # Even though we were searching for a 'FUSED' wwpn, the mapping itself
        # will have nothing on it, to indicate that the API should generate
        # the WWPNs.  Therefore, we validate that we found the mapping without
        # any WWPNs on it.
        self.assertEqual([], maps[0].client_adapter.wwpns)

        # Now try to remove it...
        vfc_mapper.remove_maps(v_wrap, self.lpar_uuid, port_map=p_map_vio1)
        self.assertEqual(len_before, len(v_wrap.vfc_mappings))

    def test_add_map(self):
        """Validates the add_map method."""
        # Determine the vios original values
        vios_wrap = self.entries[0]
        vios1_orig_map_count = len(vios_wrap.vfc_mappings)

        # Subset the WWPNs on that VIOS
        fabric_wwpns = ['10000090FA5371F2']

        # Fake Virtual WWPNs
        v_fabric_wwpns = ['0', '1']

        # Get the mappings
        fabric_map = vfc_mapper.derive_npiv_map([vios_wrap], fabric_wwpns,
                                                v_fabric_wwpns)[0]

        # Make sure the map was not there initially.
        maps = vfc_mapper.find_maps(vios_wrap.vfc_mappings, self.lpar_uuid,
                                    port_map=fabric_map)
        self.assertEqual(0, len(maps))

        # Now call the add action
        resp = vfc_mapper.add_map(vios_wrap, 'host_uuid', self.lpar_uuid,
                                  fabric_map)
        self.assertIsNotNone(resp)
        self.assertIsInstance(resp, pvm_vios.VFCMapping)

        # Verify the update is now found.
        maps = vfc_mapper.find_maps(vios_wrap.vfc_mappings, self.lpar_uuid,
                                    port_map=fabric_map)
        self.assertEqual(1, len(maps))
        self.assertEqual(vios1_orig_map_count + 1, len(vios_wrap.vfc_mappings))

        # Try to add it again...it shouldn't re-add it because its already
        # there.  Flip WWPNs to verify set query.
        fabric_map = ('10000090FA5371F2', '1 0')
        resp = vfc_mapper.add_map(vios_wrap, 'host_uuid', self.lpar_uuid,
                                  fabric_map)
        self.assertIsNone(resp)
        self.assertEqual(vios1_orig_map_count + 1, len(vios_wrap.vfc_mappings))

        # We should only find one here...the original add.  Not two even though
        # we've called add twice.
        maps = vfc_mapper.find_maps(vios_wrap.vfc_mappings, self.lpar_uuid,
                                    port_map=fabric_map)
        self.assertEqual(1, len(maps))

        # This time, remove the backing port of the existing mapping and try
        # the add again. It should return an updated mapping that contains the
        # backing port. This simulates a VM migrating with a vfc mapping, but
        # no volume had been previously detached.
        maps[0].element.remove(maps[0].backing_port.element)
        resp = vfc_mapper.add_map(vios_wrap, 'host_uuid', self.lpar_uuid,
                                  fabric_map)
        self.assertIsNotNone(resp)
        self.assertIsInstance(resp, pvm_vios.VFCMapping)
        self.assertIsNotNone(resp.backing_port)
        self.assertIn('Port', resp.child_order)

        # Pass in slot number to be set on the VFC adapter
        fabric_map = ('10000090FA5371F1', '2 3')
        resp = vfc_mapper.add_map(vios_wrap, 'host_uuid', self.lpar_uuid,
                                  fabric_map, lpar_slot_num=3)
        self.assertIsNotNone(resp)
        self.assertEqual(vios1_orig_map_count + 2, len(vios_wrap.vfc_mappings))
        # Verify the update is now found.
        maps = vfc_mapper.find_maps(vios_wrap.vfc_mappings, self.lpar_uuid,
                                    port_map=fabric_map)
        self.assertEqual(1, len(maps))
        self.assertEqual(3, maps[0].client_adapter.lpar_slot_num)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from pypowervm import adapter as adp
from pypowervm import exceptions as exc
from pypowervm.tasks import cna
from pypowervm.tests import test_fixtures as fx
from pypowervm.tests.test_utils import test_wrapper_abc as twrap
from pypowervm.wrappers import entry_wrapper as ewrap
from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import network as pvm_net

VSWITCH_FILE = 'fake_vswitch_feed.txt'
VNET_FILE = 'fake_virtual_network_feed.txt'


class TestCNA(twrap.TestWrapper):
    """Unit Tests for creating Client Network Adapters."""
    mock_adapter_fx_args = {'traits': fx.RemoteHMCTraits}
    file = VSWITCH_FILE
    wrapper_class_to_test = pvm_net.VSwitch

    @mock.patch('pypowervm.tasks.cna._find_or_create_vnet')
    def test_crt_cna(self, mock_vnet_find):
        """Tests the creation of Client Network Adapters."""
        # Create a side effect that can validate the input into the create
        # call.
        def validate_of_create(*kargs, **kwargs):
            self.assertIsNotNone(kargs[0])
            self.assertEqual('LogicalPartition', kargs[1])
            self.assertEqual('fake_lpar', kwargs.get('root_id'))
            self.assertEqual('ClientNetworkAdapter', kwargs.get('child_type'))
            return pvm_net.CNA.bld(self.adpt, 1, 'href').entry
        self.adpt.create.side_effect = validate_of_create
        self.adpt.read.return_value = self.resp

        n_cna = cna.crt_cna(self.adpt, 'fake_host', 'fake_lpar', 5)
        self.assertIsNotNone(n_cna)
        self.assertIsInstance(n_cna, pvm_net.CNA)
        self.assertEqual(1, mock_vnet_find.call_count)

    @mock.patch('pypowervm.tasks.cna._find_or_create_vnet')
    def test_crt_cna_no_vnet_crt(self, mock_vnet_find):
        """Tests the creation of Client Network Adapters.

        The virtual network creation shouldn't be done in this flow.
        """
        # PVMish Traits
        self.adptfx.set_traits(fx.LocalPVMTraits)

        self.adpt.read.return_value = self.resp

        # Create a side effect that can validate the input into the create
        # call.
        def validate_of_create(*kargs, **kwargs):
            self.assertIsNotNone(kargs[0])
            self.assertEqual('LogicalPartition', kargs[1])
            self.assertEqual('fake_lpar', kwargs.get('root_id'))
            self.assertEqual('ClientNetworkAdapter', kwargs.get('child_type'))
            return pvm_net.CNA.bld(self.adpt, 1, 'href').entry
        self.adpt.create.side_effect = validate_of_create

        n_cna = cna.crt_cna(self.adpt, 'fake_host', 'fake_lpar', 5, slot_num=1)
        self.assertIsNotNone(n_cna)
        self.assertIsInstance(n_cna, pvm_net.CNA)
        self.assertEqual(0, mock_vnet_find.call_count)

    def test_find_or_create_vswitch(self):
        """Validates that a vswitch can be created."""
        self.adpt.read.return_value = self.resp
        # Test that it finds the right vSwitch
        vswitch_w = cna._find_or_create_vswitch(self.adpt, 'fake_host',
                                                'ETHERNET0', True)
        self.assertIsNotNone(vswitch_w)

        # Create a side effect that can validate the input into the create call
        def validate_of_create(*kargs, **kwargs):
            self.assertIsNotNone(kargs[0])
            # Is the vSwitch create
            self.assertEqual('ManagedSystem', kargs[1])
            self.assertEqual('VirtualSwitch', kwargs.get('child_type'))
            # Return a previously created vSwitch...
            return self.dwrap.entry
        self.adpt.create.side_effect = validate_of_create

        # Test the create
        vswitch_w = cna._find_or_create_vswitch(self.adpt, 'fake_host',
                                                'Temp', True)
        self.assertIsNotNone(vswitch_w)
        self.assertTrue(self.adpt.create.called)

        # Make sure that if the create flag is set to false, an error is thrown
        # when the vswitch can't be found.
        self.assertRaises(exc.Error, cna._find_or_create_vswitch, self.adpt,
                          'fake_host', 'Temp', False)


class TestVNET(twrap.TestWrapper):
    mock_adapter_fx_args = {'traits': fx.RemoteHMCTraits}
    file = VNET_FILE
    wrapper_class_to_test = pvm_net.VNet

    def test_find_or_create_vnet(self):
        """Tests that the virtual network can be found/created."""
        self.adpt.read.return_value = self.resp
        fake_vs = mock.Mock()
        fake_vs.switch_id = 0
        fake_vs.name = 'ETHERNET0'

        host_uuid = '67dca605-3923-34da-bd8f-26a378fc817f'
        fake_vs.related_href = ('https://9.1.2.3:12443/rest/api/uom/'
                                'ManagedSystem/'
                                '67dca605-3923-34da-bd8f-26a378fc817f/'
                                'VirtualSwitch/'
                                'ec8aaa54-9837-3c23-a541-a4e4be3ae489')

        # This should find a vnet.
        vnet_resp = cna._find_or_create_vnet(self.adpt, host_uuid, '2227',
                                             fake_vs)
        self.assertIsNotNone(vnet_resp)

        # Now flip to a CNA that requires a create...
        resp = adp.Response('reqmethod', 'reqpath', 'status', 'reason', {})
        resp.entry = ewrap.EntryWrapper._bld(
            self.adpt, tag='VirtualNetwork').entry
        self.adpt.create.return_value = resp
        vnet_resp = cna._find_or_create_vnet(self.adpt, host_uuid, '2228',
                                             fake_vs)
        self.assertIsNotNone(vnet_resp)
        self.assertEqual(1, self.adpt.create.call_count)

    def test_find_free_vlan(self):
        """Tests that a free VLAN can be found."""
        self.adpt.read.return_value = self.resp
        # Mock data specific to the VNET File
        host_uuid = '67dca605-3923-34da-bd8f-26a378fc817f'
        fake_vs = mock.Mock()
        fake_vs.name = 'ETHERNET0'
        fake_vs.related_href = ('https://9.1.2.3:12443/rest/api/uom/'
                                'ManagedSystem/'
                                '67dca605-3923-34da-bd8f-26a378fc817f/'
                                'VirtualSwitch/'
                                'ec8aaa54-9837-3c23-a541-a4e4be3ae489')

        self.assertEqual(1, cna._find_free_vlan(self.adpt, host_uuid, fake_vs))

    @mock.patch('pypowervm.wrappers.network.VNet.wrap')
    def test_find_free_vlan_mocked(self, mock_vnet_wrap):
        """Uses lots of mock data for a find vlan."""
        self.adpt.read.return_value = mock.Mock()

        # Helper function to build the vnets.
        def build_mock_vnets(max_vlan, vswitch_uri):
            vnets = []
            for x in range(1, max_vlan + 1):
                vnets.append(mock.Mock(vlan=x,
                                       associated_switch_uri=vswitch_uri))
            return vnets

        mock_vswitch = mock.Mock(related_href='test_vs')

        # Test when all the vnet's are on a single switch.
        mock_vnet_wrap.return_value = build_mock_vnets(3000, 'test_vs')
        self.assertEqual(3001, cna._find_free_vlan(self.adpt, 'host_uuid',
                                                   mock_vswitch))

        # Test with multiple switches.  The second vswitch with a higher vlan
        # should not impact the vswitch we're searching for.
        mock_vnet_wrap.return_value = (build_mock_vnets(2000, 'test_vs') +
                                       build_mock_vnets(4000, 'test_vs2'))
        self.assertEqual(2001, cna._find_free_vlan(self.adpt, 'host_uuid',
                                                   mock_vswitch))

        # Test when all the VLANs are consumed
        mock_vnet_wrap.return_value = build_mock_vnets(4094, 'test_vs')
        self.assertRaises(exc.Error, cna._find_free_vlan, self.adpt,
                          'host_uuid', mock_vswitch)

    @mock.patch('pypowervm.tasks.cna._find_free_vlan')
    def test_assign_free_vlan(self, mock_find_vlan):
        mock_find_vlan.return_value = 2016
        mocked = mock.MagicMock()
        mock_cna = mock.MagicMock(pvid=31, enabled=False)
        mock_cna.update.return_value = mock_cna
        updated_cna = cna.assign_free_vlan(mocked, mocked, mocked, mock_cna)
        self.assertEqual(2016, updated_cna.pvid)
        self.assertEqual(mock_cna.enabled, updated_cna.enabled)
        updated_cna = cna.assign_free_vlan(mocked, mocked, mocked, mock_cna,
                                           ensure_enabled=True)
        self.assertEqual(True, updated_cna.enabled)

    @mock.patch('pypowervm.wrappers.network.CNA.bld')
    @mock.patch('pypowervm.tasks.cna._find_free_vlan')
    @mock.patch('pypowervm.tasks.cna._find_or_create_vswitch')
    @mock.patch('pypowervm.tasks.partition.get_partitions')
    def test_crt_p2p_cna(
            self, mock_get_partitions, mock_find_or_create_vswitch,
            mock_find_free_vlan, mock_cna_bld):
        """Tests the crt_p2p_cna."""
        # Mock out the data
        mock_vswitch = mock.Mock(related_href='vswitch_href')
        mock_find_or_create_vswitch.return_value = mock_vswitch
        mock_find_free_vlan.return_value = 2050

        # Mock the get of the VIOSes
        mock_vio1 = mock.Mock(uuid='src_io_host_uuid')
        mock_vio2 = mock.Mock(uuid='vios_uuid2')
        mock_get_partitions.return_value = [mock_vio1, mock_vio2]

        mock_cna = mock.MagicMock()
        mock_trunk1, mock_trunk2 = mock.MagicMock(pvid=2050), mock.MagicMock()
        mock_trunk1.create.return_value = mock_trunk1
        mock_cna_bld.side_effect = [mock_trunk1, mock_trunk2, mock_cna]

        # Invoke the create
        client_adpt, trunk_adpts = cna.crt_p2p_cna(
            self.adpt, 'host_uuid', 'lpar_uuid',
            ['src_io_host_uuid', 'vios_uuid2'], mock_vswitch, crt_vswitch=True,
            slot_num=1, mac_addr='aabbccddeeff')

        # Make sure the client and trunk were 'built'
        mock_cna_bld.assert_any_call(self.adpt, 2050, 'vswitch_href',
                                     slot_num=1, mac_addr='aabbccddeeff')
        mock_cna_bld.assert_any_call(self.adpt, 2050, 'vswitch_href',
                                     trunk_pri=1, dev_name=None)
        mock_cna_bld.assert_any_call(self.adpt, 2050, 'vswitch_href',
                                     trunk_pri=2, dev_name=None)

        # Make sure they were then created
        self.assertIsNotNone(client_adpt)
        self.assertEqual(2, len(trunk_adpts))
        mock_cna.create.assert_called_once_with(
            parent_type=pvm_lpar.LPAR, parent_uuid='lpar_uuid')
        mock_trunk1.create.assert_called_once_with(parent=mock_vio1)
        mock_trunk2.create.assert_called_once_with(parent=mock_vio2)

    @mock.patch('pypowervm.wrappers.network.CNA.bld')
    @mock.patch('pypowervm.tasks.cna._find_free_vlan')
    @mock.patch('pypowervm.tasks.cna._find_or_create_vswitch')
    @mock.patch('pypowervm.tasks.partition.get_partitions')
    def test_crt_p2p_cna_single(
            self, mock_get_partitions, mock_find_or_create_vswitch,
            mock_find_free_vlan, mock_cna_bld):
        """Tests the crt_p2p_cna with the mgmt lpar and a dev_name."""
        # Mock out the data
        mock_vswitch = mock.Mock(related_href='vswitch_href')
        mock_find_or_create_vswitch.return_value = mock_vswitch
        mock_find_free_vlan.return_value = 2050

        # Mock the get of the VIOSes
        mock_vio1 = mock.Mock(uuid='mgmt_lpar_uuid')
        mock_vio2 = mock.Mock(uuid='vios_uuid2')
        mock_get_partitions.return_value = [mock_vio1, mock_vio2]

        mock_cna = mock.MagicMock()
        mock_trunk1 = mock.MagicMock(pvid=2050)
        mock_trunk1.create.return_value = mock_trunk1
        mock_cna_bld.side_effect = [mock_trunk1, mock_cna]

        # Invoke the create
        client_adpt, trunk_adpts = cna.crt_p2p_cna(
            self.adpt, 'host_uuid', 'lpar_uuid',
            ['mgmt_lpar_uuid'], mock_vswitch, crt_vswitch=True,
            mac_addr='aabbccddeeff', dev_name='tap-12345')

        # Make sure the client and trunk were 'built'
        mock_cna_bld.assert_any_call(self.adpt, 2050, 'vswitch_href',
                                     mac_addr='aabbccddeeff', slot_num=None)
        mock_cna_bld.assert_any_call(self.adpt, 2050, 'vswitch_href',
                                     trunk_pri=1, dev_name='tap-12345')

        # Make sure they were then created
        self.assertIsNotNone(client_adpt)
        self.assertEqual(1, len(trunk_adpts))
        mock_cna.create.assert_called_once_with(
            parent_type=pvm_lpar.LPAR, parent_uuid='lpar_uuid')
        mock_trunk1.create.assert_called_once_with(parent=mock_vio1)

    @mock.patch('pypowervm.wrappers.network.CNA.bld')
    @mock.patch('pypowervm.tasks.cna._find_free_vlan')
    @mock.patch('pypowervm.tasks.cna._find_or_create_vswitch')
    @mock.patch('pypowervm.tasks.partition.get_partitions')
    def test_crt_trunk_with_free_vlan(
            self, mock_get_partitions, mock_find_or_create_vswitch,
            mock_find_free_vlan, mock_cna_bld):
        """Tests the crt_trunk_with_free_vlan on mgmt based VIOS."""
        # Mock out the data
        mock_vswitch = mock.Mock(related_href='vswitch_href')
        mock_find_or_create_vswitch.return_value = mock_vswitch
        mock_find_free_vlan.return_value = 2050

        # Mock the get of the VIOSes.
        mock_vio1 = mock.Mock(uuid='vios_uuid1')
        mock_get_partitions.return_value = [mock_vio1]

        mock_trunk1 = mock.MagicMock(pvid=2050)
        mock_trunk1.create.return_value = mock_trunk1
        mock_cna_bld.return_value = mock_trunk1

        # Invoke the create
        trunk_adpts = cna.crt_trunk_with_free_vlan(
            self.adpt, 'host_uuid', ['vios_uuid1'],
            mock_vswitch, crt_vswitch=True, dev_name='tap-12345')

        # Make sure the client and trunk were 'built'
        mock_cna_bld.assert_any_call(self.adpt, 2050, 'vswitch_href',
                                     trunk_pri=1, dev_name='tap-12345')

        # Make sure that the trunk was created
        self.assertEqual(1, len(trunk_adpts))
        mock_trunk1.create.assert_called_once_with(parent=mock_vio1)

    @mock.patch('pypowervm.wrappers.network.CNA.get')
    def test_find_trunk_on_lpar(self, mock_cna_get):
        parent_wrap = mock.MagicMock()

        m1 = mock.Mock(is_trunk=True, pvid=2, vswitch_id=2)
        m2 = mock.Mock(is_trunk=False, pvid=3, vswitch_id=2)
        m3 = mock.Mock(is_trunk=True, pvid=3, vswitch_id=1)
        m4 = mock.Mock(is_trunk=True, pvid=3, vswitch_id=2)

        mock_cna_get.return_value = [m1, m2, m3]
        self.assertIsNone(cna._find_trunk_on_lpar(self.adpt, parent_wrap, m4))
        self.assertTrue(mock_cna_get.called)

        mock_cna_get.reset_mock()
        mock_cna_get.return_value = [m1, m2, m3, m4]
        self.assertEqual(m4, cna._find_trunk_on_lpar(self.adpt, parent_wrap,
                                                     m4))
        self.assertTrue(mock_cna_get.called)

    @mock.patch('pypowervm.tasks.cna._find_trunk_on_lpar')
    @mock.patch('pypowervm.tasks.partition.get_mgmt_partition')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
    def test_find_trunks(self, mock_vios_get, mock_get_mgmt,
                         mock_find_trunk):
        # Mocked responses can be simple, since they are just fed into the
        # _find_trunk_on_lpar
        mock_vios_get.return_value = [mock.MagicMock(), mock.MagicMock()]
        mock_get_mgmt.return_value = mock.MagicMock()

        # The responses back from the find trunk.  Make it an odd trunk
        # priority ordering to make sure we sort properly
        v1 = mock.Mock(trunk_pri=3)
        c1, c2 = mock.Mock(trunk_pri=1), mock.Mock(trunk_pri=2)
        mock_find_trunk.side_effect = [v1, c1, c2]

        # Invoke the method.
        resp = cna.find_trunks(self.adpt, mock.Mock(pvid=2))

        # Make sure four calls to the find trunk
        self.assertEqual(3, mock_find_trunk.call_count)

        # Order of the response is important.  Should be based off of trunk
        # priority
        self.assertEqual([c1, c2, v1], resp)

    @mock.patch('pypowervm.wrappers.network.CNA.get')
    def test_find_all_trunks_on_lpar(self, mock_cna_get):
        parent_wrap = mock.MagicMock()

        m1 = mock.Mock(is_trunk=True, vswitch_id=2)
        m2 = mock.Mock(is_trunk=False, vswitch_id=2)
        m3 = mock.Mock(is_trunk=True, vswitch_id=1)
        m4 = mock.Mock(is_trunk=True, vswitch_id=2)

        mock_cna_get.return_value = [m1, m2, m3, m4]
        returnVal = [m1, m3, m4]
        self.assertEqual(returnVal, cna._find_all_trunks_on_lpar(self.adpt,
                                                                 parent_wrap))

        mock_cna_get.reset_mock()
        mock_cna_get.return_value = [m1, m2, m3, m4]
        self.assertEqual([m3],
                         cna._find_all_trunks_on_lpar(self.adpt,
                                                      parent_wrap=parent_wrap,
                                                      vswitch_id=1))

    @mock.patch('pypowervm.wrappers.network.CNA.get')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.get')
    def test_find_cna_wraps(self, mock_lpar_get, mock_vios_get, mock_cna_get):
        # Mocked responses are simple since they are only used for
        # pvm_net.CNA.get
        mock_lpar_get.return_value = [mock.MagicMock()]
        mock_vios_get.return_value = [mock.MagicMock()]

        # Mocked cna_wraps
        m1 = mock.Mock(uuid=2, pvid=2, vswitch_id=2)
        m2 = mock.Mock(uuid=3, pvid=1, vswitch_id=1)
        m3 = mock.Mock(uuid=1, pvid=1, vswitch_id=1)

        mock_cna_get.side_effect = [[m1, m2], [m3]]
        mock_trunk = mock.Mock(adapter=self.adpt, uuid=1, pvid=1, vswitch_id=1)
        self.assertEqual([m1, m2, m3], cna._find_cna_wraps(mock_trunk))

        mock_cna_get.side_effect = [[m1, m2], [m3]]
        self.assertEqual([m2, m3], cna._find_cna_wraps(mock_trunk, 1))

    @mock.patch('pypowervm.tasks.cna._find_cna_wraps')
    def test_find_cnas_on_trunk(self, mock_find_wraps):

        # Mocked cna_wraps
        m1 = mock.Mock(uuid=2, pvid=2, vswitch_id=2)
        m2 = mock.Mock(uuid=3, pvid=1, vswitch_id=1)
        m3 = mock.Mock(uuid=1, pvid=1, vswitch_id=1)

        mock_find_wraps.return_value = [m1, m2, m3]
        mock_trunk = mock.Mock(adapter=self.adpt, uuid=1, pvid=1, vswitch_id=1)
        self.assertEqual([m2], cna.find_cnas_on_trunk(mock_trunk))

        mock_find_wraps.return_value = [m1, m3]
        self.assertEqual([], cna.find_cnas_on_trunk(mock_trunk))
        mock_trunk = mock.Mock(adapter=self.adpt, uuid=3, pvid=3, vswitch_id=3)
        self.assertEqual([], cna.find_cnas_on_trunk(mock_trunk))

    @mock.patch('pypowervm.tasks.cna._find_cna_wraps')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
    @mock.patch('pypowervm.tasks.partition.get_mgmt_partition')
    @mock.patch('pypowervm.tasks.cna._find_all_trunks_on_lpar')
    @mock.patch('pypowervm.wrappers.network.VSwitch.search')
    def test_find_orphaned_trunks(self, mock_vswitch, mock_trunks,
                                  mock_get_mgmt, mock_vios_get, mock_wraps):

        mock_vswitch.return_value = mock.MagicMock(switch_id=1)
        mock_get_mgmt.return_value = mock.MagicMock()
        mock_vios_get.return_value = [mock.MagicMock()]
        # Mocked cna_wraps
        m1 = mock.Mock(is_trunk=True, uuid=2, pvid=2, vswitch_id=1)
        m2 = mock.Mock(is_trunk=False, uuid=3, pvid=3, vswitch_id=1)
        m3 = mock.Mock(is_trunk=True, uuid=1, pvid=1, vswitch_id=1)
        m4 = mock.Mock(is_trunk=False, uuid=4, pvid=1, vswitch_id=1)
        mock_wraps.return_value = [m1, m2, m3, m4]
        mock_trunks.side_effect = [[m1, m3], []]

        self.assertEqual([m1], cna.find_orphaned_trunks(self.adpt,
                                                        mock.MagicMock))



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tasks to request and release master mode."""

from oslo_config import cfg
from oslo_log import log as logging

import pypowervm.const as c
import pypowervm.log as lgc
from pypowervm.wrappers import job
import pypowervm.wrappers.managed_system as ms

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

_SUFFIX_PARM_REQUEST_MASTER = 'RequestMaster'
_SUFFIX_PARM_RELEASE_MASTER = 'ReleaseMaster'

CO_MGMT_MASTER_STATUS = "coManagementMasterStatus"


class MasterMode(object):
    """Valid master modes used when requesting master.

    NORMAL: Default mode
    TEMP:   When released, the original master is immediately restored.
    """
    NORMAL = "norm"
    TEMP = "temp"


@lgc.logcall
def request_master(msys, mode=MasterMode.NORMAL,
                   timeout=CONF.pypowervm_job_request_timeout):
    """Request master mode for the provided Managed System.

    :param msys: Managed System wrapper requesting master mode
    :param mode: The requested master mode type.
                 There are 2 options:
                     MasterMode.NORMAL ("norm"): default
                     MasterMode.TEMP ("temp"): when released, the original
                                               master is immediately restored
    :param timeout: maximum number of seconds for job to complete
    """
    resp = msys.adapter.read(ms.System.schema_type, msys.uuid,
                             suffix_type=c.SUFFIX_TYPE_DO,
                             suffix_parm=_SUFFIX_PARM_REQUEST_MASTER)
    job_wrapper = job.Job.wrap(resp.entry)
    job_parms = [job_wrapper.create_job_parameter(CO_MGMT_MASTER_STATUS,
                                                  mode)]
    job_wrapper.run_job(msys.uuid, job_parms=job_parms, timeout=timeout)


@lgc.logcall
def release_master(msys, timeout=CONF.pypowervm_job_request_timeout):
    """Release master mode for the provided Managed System.

    :param msys: Managed System wrapper requesting master mode
    :param timeout: maximum number of seconds for job to complete
    """
    resp = msys.adapter.read(ms.System.schema_type, msys.uuid,
                             suffix_type=c.SUFFIX_TYPE_DO,
                             suffix_parm=_SUFFIX_PARM_RELEASE_MASTER)
    job_wrapper = job.Job.wrap(resp.entry)
    job_wrapper.run_job(msys.uuid, timeout=timeout)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import os

import pypowervm.adapter as adp

EOL = "\n"

COMMENT = "#"
INFO = "INFO{"
HEADERS = "HEADERS{"
BODY = "BODY{"
END_OF_SECTION = "END OF SECTION}"


class PVMFile(object):
    def __init__(self, file_name=None):
        self.comment = None
        self.path = None
        self.reason = None
        self.status = None
        self.headers = None
        self.body = None

        if file_name is not None:
            self.load_file(file_name)

    def load_file(self, file_name):
        """Load a REST response file."""
        # If given a pathed filename, use it
        dirname = os.path.dirname(file_name)
        if not dirname:
            dirname = os.path.dirname(os.path.dirname(__file__))
            file_name = os.path.join(dirname, "data", file_name)

        resp_file = open(file_name, "r")

        if resp_file is None:
            raise Exception("Could not load %s" % file_name)

        while True:
            line = resp_file.readline()
            if line is None or len(line) == 0:
                break

            if len(line.strip()) == 0:
                continue

            if line.startswith(COMMENT):
                continue

            if line.startswith(INFO):
                section = INFO
            elif line.startswith(HEADERS):
                section = HEADERS
            elif line.startswith(BODY):
                section = BODY
            else:
                resp_file.close()
                raise Exception("Unknown line in file %s: %s" %
                                (file_name, line))

            buf = _read_section(section, file_name, resp_file)

            if line.startswith(INFO):
                info = ast.literal_eval(buf)
                self.comment = info['comment']
                self.path = info['path']
                self.reason = info['reason']
                self.status = info['status']
            elif line.startswith(HEADERS):
                self.headers = ast.literal_eval(buf)
            elif line.startswith(BODY):
                self.body = buf

        resp_file.close()


class PVMResp(PVMFile):

    """Class to encapsulate the text serialization of a response."""

    def __init__(self, file_name=None, pvmfile=None, adapter=None):
        """Initialize this PVMResp by loading a file or pulling a PVMFile.

        :param file_name: Name of a file to load.
        :param pvmfile: An existing PVMFile instance.  This PVMResp will use
                        its attributes.  If both file_name and pvmfile are
                        specified, the file will be reloaded into the passed-in
                        PVMFile.  This is probably not what you intended.
        :param adapter: A pypowervm.adapter.Adapter, used for traits, etc.
        """
        super(PVMResp, self).__init__()
        # Legacy no-arg constructor - allow caller to set fields manually
        if pvmfile is None and file_name is None:
            return
        if pvmfile is None:
            self.load_file(file_name)
        else:
            # Use pvmfile
            if file_name is not None:
                pvmfile.load_file(file_name)
            # Copy in attrs from pvmfile
            self.comment = pvmfile.comment
            self.path = pvmfile.path
            self.reason = pvmfile.reason
            self.status = pvmfile.status
            self.headers = pvmfile.headers
            self.body = pvmfile.body

        self.response = adp.Response(reqmethod=None, reqpath=None,
                                     status=self.status, reason=self.reason,
                                     headers=self.headers, body=self.body)
        self.response.adapter = adapter
        self.response._unmarshal_atom()

    def get_response(self):
        return self.response

    def refresh(self):
        """Do the query and get the response."""

        print("Connecting.")
        adap = adp.Adapter()

        print("Reading path:  " + self.path)
        self.response = adap.read(self.path)

        print("Received " + str(self.response))

    def save(self, file_name):

        everything = {
            'comment': self.comment,
            'path': self.path,
            'reason': self.response.reason,
            'status': self.response.status,
        }

        with open(file_name, 'wb') as df:
            df.write("####################################################")
            df.write(EOL)
            df.write("# THIS IS AN AUTOMATICALLY GENERATED FILE")
            df.write(EOL)
            df.write("# DO NOT EDIT. ANY EDITS WILL BE LOST ON NEXT UPDATE")
            df.write(EOL)
            df.write("#")
            df.write(EOL)
            df.write("# To update file, run: create_httpresp.py -refresh ")
            df.write(os.path.basename(file_name))
            df.write(EOL)
            df.write("#")
            df.write(EOL)
            df.write("####################################################")
            df.write(EOL)

            df.write(INFO + EOL)
            df.write(str(everything))
            df.write(EOL)
            df.write(END_OF_SECTION)
            df.write(EOL)
            df.write(HEADERS + EOL)
            df.write(str(self.response.headers))
            df.write(EOL)
            df.write(END_OF_SECTION)
            df.write(EOL)
            df.write(BODY + EOL)
            df.write(self.response.body)
            df.write(EOL)
            df.write(END_OF_SECTION)
            df.write(EOL)


def load_pvm_resp(file_name, adapter=None):
    return PVMResp(file_name, adapter=adapter)


def _read_section(section, file_name, resp_file):

    buf = ""
    while True:
        line = resp_file.readline()
        if line is None or len(line) == 0:
            raise Exception("Could not find end of section %s of file %s" %
                            (section, file_name))

        if line.startswith(END_OF_SECTION):
            return buf

        buf += EOL + line



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""EntryWrapper, constants, and enums around Job ('web' namespace)."""

import threading
import time

from oslo_config import cfg
from oslo_log import log as logging
import six

import pypowervm.const as pc
import pypowervm.entities as ent
import pypowervm.exceptions as pvmex
from pypowervm.i18n import _
import pypowervm.util as u
import pypowervm.wrappers.entry_wrapper as ewrap

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

_JOBS = 'jobs'
_REQ_OP = 'RequestedOperation'
_JOB_GROUP_NAME = u.xpath(_REQ_OP, 'GroupName')
_JOB_OPERATION_NAME = u.xpath(_REQ_OP, 'OperationName')
_JOB_PARAM = u.xpath('Results', 'JobParameter')
_JOB_RESULTS_NAME = u.xpath(_JOB_PARAM, 'ParameterName')
_JOB_RESULTS_VALUE = u.xpath(_JOB_PARAM, 'ParameterValue')
_RESPONSE_EXCEPTION = 'ResponseException'
_JOB_MESSAGE = u.xpath(_RESPONSE_EXCEPTION, 'Message')
_JOB_STACKTRACE = u.xpath(_RESPONSE_EXCEPTION, 'StackTrace')
_JOB_STATUS = 'Status'
_JOB_ID = 'JobID'


class JobStatus(object):
    NOT_ACTIVE = 'NOT_STARTED'
    RUNNING = 'RUNNING'
    COMPLETED_OK = 'COMPLETED_OK'
    COMPLETED_WITH_WARNINGS = 'COMPLETED_WITH_WARNINGS'
    COMPLETED_WITH_ERROR = 'COMPLETED_WITH_ERROR'


class PollAndDeleteThread(threading.Thread):
    def __init__(self, job, sensitive):
        super(PollAndDeleteThread, self).__init__()
        self.job = job
        self.sensitive = sensitive

    def run(self):
        self.job.poll_while_status([JobStatus.RUNNING], 0, self.sensitive)
        self.job.delete_job()
        # If the Job failed, we still want to log it.
        if self.job.job_status != JobStatus.COMPLETED_OK:
            exc = pvmex.JobRequestFailed(
                operation_name=self.job.op, error=self.job.get_job_message())
            LOG.error(exc.args[0])


class CancelJobThread(threading.Thread):
    def __init__(self, job, sensitive):
        super(CancelJobThread, self).__init__()
        self.job = job
        self.sensitive = sensitive

    def run(self):
        self.job._monitor_job(timeout=0, sensitive=self.sensitive)
        self.job.delete_job()


@ewrap.EntryWrapper.pvm_type('Job', ns=pc.WEB_NS)
class Job(ewrap.EntryWrapper):
    """Wrapper object for job response schema."""

    @classmethod
    def wrap(cls, response_or_entry, etag=None):
        wrap = super(Job, cls).wrap(response_or_entry, etag=etag)
        wrap.op = wrap._get_val_str(_JOB_OPERATION_NAME)
        return wrap

    @staticmethod
    def create_job_parameter(name, value, cdata=False):
        """Creates a JobParameter Element.

           :param name: ParameterName text value
           :param value: ParameterValue text value
           :param cdata: If True, the value text will be wrapped in CDATA tags
           :returns: JobParameter Element
        """
        # JobParameter doesn't need adapter today
        adapter = None

        job_parm = ent.Element('JobParameter', adapter,
                               attrib={'schemaVersion': 'V1_0'},
                               ns=pc.WEB_NS)
        job_parm.append(ent.Element('ParameterName', adapter,
                                    text=name, ns=pc.WEB_NS))
        job_parm.append(ent.Element('ParameterValue', adapter,
                                    text=value, ns=pc.WEB_NS, cdata=cdata))
        return job_parm

    def add_job_parameters_to_existing(self, *add_parms):
        """Adds JobParameter Elements to existing JobParameters xml.

           Must be a job response entry.
           :param add_parms: list of JobParamters to add
        """
        job_parms = self.entry.element.find('JobParameters')
        for parm in add_parms:
            job_parms.append(parm)

    @property
    def job_id(self):
        """Gets the job ID string.

        :returns: String containing the job ID
        """
        return self._get_val_str(_JOB_ID)

    @property
    def job_status(self):
        """Gets the job status string.

        :returns: String containing the job status
        """
        return self._get_val_str(_JOB_STATUS)

    def get_job_resp_exception_msg(self, default=''):
        """Gets the job message string from the ResponseException.

        :returns: String containing the job message or
                  default (defaults to empty string) if not found
        """
        job_message = self._get_val_str(_JOB_MESSAGE, default)
        if job_message:
            # See if there is a stack trace to log
            stack_trace = self._get_val_str(_JOB_STACKTRACE, default)
            if stack_trace:
                LOG.error(pvmex.JobRequestFailed(operation_name=self.op,
                                                 error=stack_trace))
        return job_message

    def get_job_results_message(self, default=''):
        """Gets the job result message string.

        :returns: String containing the job result message or
                  default (defaults to empty string) if not found
        """
        message = default
        parm_names = self._get_vals(_JOB_RESULTS_NAME)
        parm_values = self._get_vals(_JOB_RESULTS_VALUE)
        for i in range(len(parm_names)):
            if parm_names[i] == 'result':
                message = parm_values[i]
                break
        return message

    def get_job_results_as_dict(self, default=None):
        """Gets the job results as a dictionary.

        :returns: Dictionary with result parm names and parm
                  values as key, value pairs.
        """
        results = default if default else {}
        parm_names = self._get_vals(_JOB_RESULTS_NAME)
        parm_values = self._get_vals(_JOB_RESULTS_VALUE)
        for i in range(len(parm_names)):
            results[parm_names[i]] = parm_values[i]
        return results

    def get_job_message(self, default=''):
        """Gets the job message string.

        It checks job results message first, if results message is not found,
        it checks for a ResponseException message. If neither is found, it
        returns the default.

        :returns: String containing the job message or
                  default (defaults to empty string) if not found
        """
        message = self.get_job_results_message(default=default)
        if not message:
            message = self.get_job_resp_exception_msg(default=default)
        return message

    def run_job(self, uuid, job_parms=None,
                timeout=CONF.pypowervm_job_request_timeout,
                sensitive=False, synchronous=True):
        """Invokes and polls a job.

        Adds job parameters to the job element if specified and calls the
        create_job method. It then monitors the job for completion and sends a
        JobRequestFailed exception if it did not complete successfully.

        :param uuid: uuid of the target
        :param job_parms: list of JobParamters to add
        :param timeout: maximum number of seconds for job to complete
        :param sensitive: If True, mask the Job payload in the logs.
        :param synchronous: If True (the default), wait for the Job to complete
                            or time out.  If False, return as soon as the Job
                            starts.  Note that this may still involve polling
                            (if the Job is waiting in queue to start), and may
                            still time out (if the Job hasn't started within
                            the requested timeout.)
        :raise JobRequestFailed: if the job did not complete successfully.
        :raise JobRequestTimedOut: if the job timed out.
        """
        if job_parms:
            self.add_job_parameters_to_existing(*job_parms)
        try:
            self.entry = self.adapter.create_job(
                self.entry.element, self._get_val_str(_JOB_GROUP_NAME), uuid,
                sensitive=sensitive).entry
        except pvmex.Error as exc:
            LOG.exception(exc)
            raise pvmex.JobRequestFailed(operation_name=self.op, error=exc)
        timed_out = self._monitor_job(
            timeout=timeout, sensitive=sensitive, synchronous=synchronous)
        if timed_out:
            try:
                self.cancel_job()
            except pvmex.JobRequestFailed as e:
                LOG.warning(six.text_type(e))
            exc = pvmex.JobRequestTimedOut(
                operation_name=self.op, seconds=timeout)
            LOG.error(exc.args[0])
            raise exc
        if not synchronous:
            # _monitor_job spawned a subthread that will delete_job when done.
            return
        self.delete_job()
        if self.job_status != JobStatus.COMPLETED_OK:
            exc = pvmex.JobRequestFailed(
                operation_name=self.op, error=self.get_job_message(''))
            LOG.error(exc.args[0])
            raise exc

    def poll_while_status(self, statuses, timeout, sensitive):
        """Poll the Job as long as its status is in the specified list.

        :param statuses: Iterable of JobStatus enum values.  This method
                         continues to poll the Job as long as its status is
                         in the specified list, or until the timeout is
                         reached (whichever comes first).
        :param timeout: Maximum number of seconds to keep checking job status.
                        If zero, poll indefinitely.
        :param sensitive: If True, mask the Job payload in the logs.
        :return: timed_out: True if the timeout was reached before the Job
                            left the specified set of states.
        """
        start_time = time.time()
        iteration_count = 1
        while self.job_status in statuses:
            elapsed_time = time.time() - start_time
            if timeout:
                # wait up to timeout seconds
                if elapsed_time > timeout:
                    return True
            # Log a warning every 5 minutes
            if not iteration_count % 300:
                msg = _("Job %(job_id)s monitoring for %(time)i seconds.")
                LOG.warning(msg, {'job_id': self.job_id, 'time': elapsed_time})
            time.sleep(1)
            self.entry = self.adapter.read_job(
                self.job_id, sensitive=sensitive).entry
            iteration_count += 1
        return False

    def _monitor_job(self, timeout=CONF.pypowervm_job_request_timeout,
                     sensitive=False, synchronous=True):
        """Polls a job.

        Waits on a job until it is no longer running.  If a timeout is given,
        it times out in the given amount of time.

        :param timeout: maximum number of seconds to keep checking job status
        :param sensitive: If True, mask the Job payload in the logs.
        :param synchronous: If True (the default), wait for the Job to complete
                            or time out.  If False, return as soon as the Job
                            starts.  Note that this may still involve polling
                            (if the Job is waiting in queue to start), and may
                            still time out (if the Job hasn't started within
                            the requested timeout.)  If synchronous=True, the
                            caller must delete the Job (self.delete_job()); if
                            False, this method spawns a subthread that deletes
                            it when it finishes.
        :returns timed_out: boolean True if timed out waiting for job
                            completion
        """
        if synchronous:
            return self.poll_while_status(
                [JobStatus.RUNNING, JobStatus.NOT_ACTIVE], timeout, sensitive)

        # Asynchronous: wait for the Job to start, then spawn a thread to wait
        # (indefinitely) for it to finish, and delete it when done.
        if self.poll_while_status([JobStatus.NOT_ACTIVE], timeout, sensitive):
            return True

        PollAndDeleteThread(self, sensitive).start()
        return False

    def cancel_job(self, sensitive=False):
        """Cancels and deletes incomplete/running jobs.

        This method spawns a thread to monitor the job being cancelled
        and delete it.

        :param sensitive: If True, payload will be hidden in the logs
        """

        job_id = self.job_id
        msg = _("Issuing cancel request for job %(job_id)s. Will poll the "
                "job indefinitely for termination.")
        LOG.warning(msg, {'job_id': job_id})
        try:
            self.adapter.update(None, None, root_type=_JOBS, root_id=job_id,
                                suffix_type='cancel')
        except pvmex.Error as exc:
            LOG.exception(exc)
        CancelJobThread(self, sensitive).start()

    def delete_job(self):
        """Cleans this Job off of the REST server, if it is completed.

        :raise JobRequestFailed: if the Job is detected to be running.
        """
        if self.job_status == JobStatus.RUNNING:
            error = (_("Job %s not deleted. Job is in running state.")
                     % self.job_id)
            LOG.error(error)
            raise pvmex.Error(error)
        try:
            self.adapter.delete(_JOBS, self.job_id)
        except pvmex.Error as exc:
            LOG.exception(exc)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from pypowervm import exceptions as pvm_ex
from pypowervm.i18n import _
from pypowervm.tasks import partition
from pypowervm.wrappers import storage as pvm_stg
from pypowervm.wrappers import virtual_io_server as pvm_vios


LOG = logging.getLogger(__name__)

_cur_vios_uuid = None
_cur_vg_uuid = None


def validate_vopt_repo_exists(
        adapter, vopt_media_volume_group='rootvg', vopt_media_rep_size=1):
    """Will ensure that the virtual optical media repository exists.

    Checks to make sure that at least one Virtual I/O Server has a virtual
    optical media repository.

    If the volume group on an I/O Server goes down (perhaps due to
    maintenance), the system will rescan to determine if there is another
    I/O Server that can host the request.

    The very first invocation may be expensive.  It may also be expensive
    to call if a Virtual I/O Server unexpectedly goes down.

    :param adapter: The pypowervm adapter.
    :param vopt_media_volume_group: (Optional, Default: rootvg) The volume
                                    group to use if the vopt media repo needs
                                    to be created.
    :param vopt_media_rep_size: (Optional, Default: 1) The size of the virtual
                                optical media (in GB) if the repo needs to be
                                created.
    :return vios_uuid: The VIOS uuid hosting the VG
    :return vg_uuid: The volume group uuid hosting the vopt.
    :raise NoMediaRepoVolumeGroupFound: Raised when there are no VIOSes that
                                        can support the virtual optical media.
    """
    # If our static variables were set, then we should validate that the
    # repo is still running.  Otherwise, we need to reset the variables
    # (as it could be down for maintenance).
    if _cur_vg_uuid is not None:
        vio_uuid = _cur_vios_uuid
        vg_uuid = _cur_vg_uuid
        try:
            vg_wrap = pvm_stg.VG.get(
                adapter, uuid=vg_uuid, parent_type=pvm_vios.VIOS,
                parent_uuid=vio_uuid)
            if vg_wrap is not None and len(vg_wrap.vmedia_repos) != 0:
                return vio_uuid, vg_uuid
        except Exception as exc:
            LOG.exception(exc)

        LOG.warning(_("An error occurred querying the virtual optical "
                      "media repository.  Attempting to re-establish "
                      "connection with a virtual optical media repository"))

    # Did not find the media repository.  Need a deeper query
    return _find_or_rebuild_vopt_repo(adapter, vopt_media_volume_group,
                                      vopt_media_rep_size)


def _find_or_rebuild_vopt_repo(adapter, vopt_media_volume_group,
                               vopt_media_rep_size):
    # If we're hitting this:
    # a) It's our first time booting up;
    # b) The previously-used Volume Group went offline (e.g. VIOS went down
    #    for maintenance); OR
    # c) The previously-used media repository disappeared.
    #
    # The next step is to create a Media Repository dynamically.
    found_vg, found_vios, conf_vg, conf_vios = _find_vopt_repo_data(
        adapter, vopt_media_volume_group)

    # If we didn't find a media repos OR an appropriate volume group, raise
    # the exception.  Since vopt_media_volume_group defaults to rootvg,
    # which is always present, this should only happen if:
    # a) No media repos exists on any VIOS we can see; AND
    # b) The user specified a non-rootvg vopt_media_volume_group; AND
    # c) The specified volume group did not exist on any VIOS.
    if found_vg is None and conf_vg is None:
        raise pvm_ex.NoMediaRepoVolumeGroupFound(
            vol_grp=vopt_media_volume_group)

    # If no media repos was found, create it.
    if found_vg is None:
        found_vg, found_vios = conf_vg, conf_vios
        vopt_repo = pvm_stg.VMediaRepos.bld(
            adapter, 'vopt', vopt_media_rep_size)
        found_vg.vmedia_repos = [vopt_repo]
        found_vg = found_vg.update()

    # At this point, we know that we've successfully found or created the
    # volume group.  Save to the static variables.
    global _cur_vg_uuid, _cur_vios_uuid
    _cur_vg_uuid = found_vg.uuid
    _cur_vios_uuid = found_vios.uuid
    return _cur_vios_uuid, _cur_vg_uuid


def _find_vopt_repo_data(adapter, vopt_media_volume_group):
    """Finds the vopt repo defaults.

    :param adapter: pypowervm adapter
    :param vopt_media_volume_group: The name of the volume group to use.
    :return found_vg: Returned if a volume group already exists with a media
                      repo within it.  Is that corresponding volume group.
    :return found_vios: Returned if a volume group already exists with a media
                        repo within it.  This is the VIOS wrapper.
    :return conf_vg: Returned if a volume group does not exist with a media
                     repo within it.  This is the volume group wrapper (as
                     defined by the vopt_media_volume_group) that the consumer
                     code should create that media repo within.
    :return conf_vios: Returned if a volume group does not exist with a media
                       repo within it.  This is the VIOS wrapper that is the
                       parent of the conf_vg
    """
    vios_wraps = partition.get_active_vioses(adapter)

    # First loop through the VIOSes and their VGs to see if a media repos
    # already exists.
    found_vg, found_vios = None, None

    # And in case we don't find the media repos, keep track of the VG on
    # which we should create it.
    conf_vg, conf_vios = None, None

    for vio_wrap in vios_wraps:
        vg_wraps = pvm_stg.VG.get(adapter, parent=vio_wrap)
        for vg_wrap in vg_wraps:
            if len(vg_wrap.vmedia_repos) != 0:
                found_vg, found_vios = vg_wrap, vio_wrap
                break
            # In case no media repos exists, save a pointer to the
            # CONFigured vopt_media_volume_group if we find it.
            if (conf_vg is None and not vio_wrap.is_mgmt_partition and
                    vg_wrap.name == vopt_media_volume_group):
                conf_vg, conf_vios = vg_wrap, vio_wrap

        # If we found it, don't keep looking
        if found_vg:
            break

    return found_vg, found_vios, conf_vg, conf_vios



# Copyright 2014, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Module-specific error/exception subclasses."""

import abc
from lxml import etree
import six

from pypowervm import const as c
from pypowervm import entities as ent
from pypowervm.i18n import _


class Error(Exception):
    """Error on PowerVM API Adapter method invocation."""

    def __init__(self, msg, response=None):
        """Initializes Error with optional `response` object."""
        self.response = response
        self.orig_response = None
        super(Error, self).__init__(msg)


class ConnectionError(Error):
    """Connection Error on PowerVM API Adapter method invocation."""


class SSLError(Error):
    """SSL Error on PowerVM API Adapter method invocation."""


class TimeoutError(Error):
    """Timeout Error on PowerVM API Adapter method invocation."""


class HttpError(Error):
    """HTTP Error on PowerVM API Adapter method invocation."""

    def __init__(self, resp):
        """Initializes HttpError with required `response` object.

        1) Constructs the exception message based on the contents of the
        response parameter.

        2) If possible, initializes a 'her_wrap' member which is a
        pypowervm.wrappers.http_error.HttpError - an EntryWrapper for the
        <HttpErrorResponse/> payload.  Consumers should check this member for
        None before using.

        :param resp: pypowervm.adapter.Response containing an
                     <HttpErrorResponse/> response from the REST server.
        """
        self.her_wrap = None
        reason = resp.reason
        # Attempt to extract PowerVM API's HttpErrorResponse object.
        # Since this is an exception path, we use best effort only - don't want
        # problems here to obscure the real exception.
        try:
            root = etree.fromstring(resp.body)
            if root is not None and root.tag == str(etree.QName(c.ATOM_NS,
                                                                'entry')):
                resp.entry = ent.Entry.unmarshal_atom_entry(root, resp)
                # Import inline to avoid circular dependencies
                import pypowervm.wrappers.http_error as he
                self.her_wrap = he.HttpError.wrap(resp)

                # Add the message to the reason if it is available.
                if self.her_wrap.message:
                    reason += ' -- ' + self.her_wrap.message
        except Exception:
            pass
        # Construct the exception message
        msg = ('HTTP error %(status)s for method %(method)s on path '
               '%(path)s: %(reason)s') % dict(status=resp.status,
                                              method=resp.reqmethod,
                                              path=resp.reqpath,
                                              reason=reason)
        # Initialize the exception
        super(HttpError, self).__init__(msg, response=resp)


class HttpNotFound(HttpError):
    """HttpError subclass where response.status == c.HTTPStatus.NOT_FOUND."""
    pass


class HttpUnauth(HttpError):
    """HttpError where response.status == c.HTTPStatus.UNAUTHORIZED."""
    pass


class AtomError(Error):
    """Atom Error on PowerVM API Adapter method invocation."""

    def __init__(self, msg, response):
        """Initializes AtomError with required `response` object."""
        super(AtomError, self).__init__(msg, response=response)


@six.add_metaclass(abc.ABCMeta)
class AbstractMsgFmtError(Error):
    """Used to raise an exception with a formattable/parameterized message.

    The subclass must set the msg_fmt class variable.  The consumer should
    instantiate the subclass with **kwargs appropriate to its msg_fmt.
    """
    def __init__(self, response=None, **kwa):
        msg = self.msg_fmt % kwa
        super(AbstractMsgFmtError, self).__init__(msg, response=response)


class UnableToDerivePhysicalPortForNPIV(AbstractMsgFmtError):
    msg_fmt = _("Unable to derive the appropriate physical FC port for WWPN "
                "%(wwpn)s.  The VIOS Extended Attribute Groups may have been "
                "insufficient.  The VIOS URI for the query was %(vio_uri)s.")


class NotFound(AbstractMsgFmtError):
    msg_fmt = _('Element not found: %(element_type)s %(element)s')


class LPARNotFound(AbstractMsgFmtError):
    msg_fmt = _('LPAR not found: %(lpar_name)s')


class JobRequestFailed(AbstractMsgFmtError):
    msg_fmt = _("The '%(operation_name)s' operation failed. %(error)s")


class JobRequestTimedOut(JobRequestFailed):
    msg_fmt = _("The '%(operation_name)s' operation failed. "
                "Failed to complete the task in %(seconds)d seconds.")


class OSShutdownNoRMC(AbstractMsgFmtError):
    msg_fmt = _("Can not perform OS shutdown on Virtual Machine %(lpar_nm)s "
                "because its RMC connection is not active.")


class VMPowerOffFailure(AbstractMsgFmtError):
    msg_fmt = _("Failed to power off Virtual Machine %(lpar_nm)s: %(reason)s")


class VMPowerOffTimeout(VMPowerOffFailure):
    msg_fmt = _("Power off of Virtual Machine %(lpar_nm)s timed out after "
                "%(timeout)d seconds.")


class VMPowerOnFailure(AbstractMsgFmtError):
    msg_fmt = _("Failed to power on Virtual Machine %(lpar_nm)s: %(reason)s")


class VMPowerOnTimeout(VMPowerOnFailure):
    msg_fmt = _("Power on of Virtual Machine %(lpar_nm)s timed out after "
                "%(timeout)d seconds.")


class PvidOfNetworkBridgeError(AbstractMsgFmtError):
    msg_fmt = _("Unable to remove VLAN %(vlan_id)d as it is the Primary VLAN "
                "Identifier on a different Network Bridge.")


class OrphanVLANFoundOnProvision(AbstractMsgFmtError):
    msg_fmt = _("Unable to provision VLAN %(vlan_id)d.  It appears to be "
                "contained on device '%(dev_name)s' on Virtual I/O Server "
                "%(vios)s.  That device is not connected to any Network "
                "Bridge (Shared Ethernet Adapter).  Please manually remove "
                "the device or add it to the Network Bridge before "
                "continuing.")


class DuplicateLUNameError(AbstractMsgFmtError):
    msg_fmt = _("A Logical Unit with name %(lu_name)s already exists on "
                "Shared Storage Pool %(ssp_name)s.")


class UnableToFindFCPortMap(AbstractMsgFmtError):
    msg_fmt = _("Unable to find a physical port to map a virtual Fibre "
                "Channel port to.  This is due to either a Virtual I/O "
                "Server being unavailable, or improper port specification "
                "for the physical Fibre Channel ports.")


class ConsoleNotLocal(AbstractMsgFmtError):
    msg_fmt = _("Unable to start the console to the Virtual Machine.  The "
                "pypowervm API is running in a non-local mode.  The console "
                "can only be deployed when pypowervm is co-located with "
                "the PowerVM API.")


class WrapperTaskNoSubtasks(AbstractMsgFmtError):
    msg_fmt = _("WrapperTask %(name)s has no subtasks!")


class FeedTaskEmptyFeed(AbstractMsgFmtError):
    msg_fmt = _("FeedTask can't have an empty feed.")


class AuthFileReadError(AbstractMsgFmtError):
    msg_fmt = _("OS denied access to file %(access_file)s.")


class AuthFileAccessError(AbstractMsgFmtError):
    msg_fmt = _("OS encountered an I/O error attempting to read file "
                "%(access_file)s: %(error)s")


class MigrationFailed(AbstractMsgFmtError):
    msg_fmt = _("The migration task failed. %(error)s")


class IBMiLoadSourceNotFound(AbstractMsgFmtError):
    msg_fmt = _("No load source found for VM %(vm_name)s")


class UnableToBuildPG83EncodingMissingParent(AbstractMsgFmtError):
    msg_fmt = _("Unable to derive the pg83 encoding for hdisk %(dev_name)s.  "
                "The parent_entry attribute is not set.  This may be due to "
                "using a PV obtained through an unsupported property chain.  "
                "The PV must be accessed via VIOS.phys_vols, VG.phys_vols, or "
                "VIOS.scsi_mappings[n].backing_storage.")


class FoundDevMultipleTimes(AbstractMsgFmtError):
    msg_fmt = _("Found device %(devname)s %(count)d times; expected to find "
                "it at most once.")


class MultipleExceptionsInFeedTask(AbstractMsgFmtError):
    """Exception concatenating messages in WrappedFailure exceptions.

    Exception raised when a pypowervm.utils.transaction.FeedTask run raises a
    tasflow.exceptions.WrappedFailure containing more than one exception.  The
    message string is a concatenation of the message strings of the wrapped
    exceptions.
    """
    def __init__(self, ft_name, wrapped_failure):
        # In case the caller wants to trap this and get at the WrappedFailure
        self.wrapped_failure = wrapped_failure
        self.msg_fmt = _("FeedTask %(ft_name)s experienced multiple "
                         "exceptions:\n\t%(concat_msgs)s")
        concat_msgs = '\n\t'.join([fail.exception_str
                                   for fail in wrapped_failure])
        super(MultipleExceptionsInFeedTask, self).__init__(
            response=None, ft_name=ft_name, concat_msgs=concat_msgs)


class ManagementPartitionNotFoundException(AbstractMsgFmtError):
    """Couldn't find exactly one management partition on the system."""
    msg_fmt = _("Expected to find exactly one management partition; found "
                "%(count)d.")


class ThisPartitionNotFoundException(AbstractMsgFmtError):
    """Couldn't find exactly one partition with the local VM's short ID."""
    msg_fmt = _("Expected to find exactly one partition with ID %(lpar_id)d; "
                "found %(count)d.")


class NoDefaultTierFoundOnSSP(AbstractMsgFmtError):
    """Looked for a default Tier on the SSP, but didn't find it."""
    msg_fmt = _("Couldn't find the default Tier on Shared Storage Pool "
                "%(ssp_name)s.")


class InvalidHostForRebuild(AbstractMsgFmtError):
    pass


class InvalidHostForRebuildNoVIOSForUDID(InvalidHostForRebuild):
    msg_fmt = _("The device with UDID %(udid)s was not found on any of the "
                "Virtual I/O Servers.")


class InvalidHostForRebuildNotEnoughVIOS(InvalidHostForRebuild):
    msg_fmt = _("There are not enough Virtual I/O Servers to support the "
                "virtual machine's device with UDID %(udid)s.")


class InvalidHostForRebuildFabricsNotFound(InvalidHostForRebuild):
    msg_fmt = _("The expected fabrics (%(fabrics)s) were not found on any of "
                "the Virtual I/O Servers.")


class InvalidHostForRebuildInvalidIOType(InvalidHostForRebuild):
    msg_fmt = _("Can not rebuild the virtual machine.  It is using an I/O "
                "type of %(io_type)s which is not supported for VM rebuild.")


class InvalidHostForRebuildSlotMismatch(InvalidHostForRebuild):
    msg_fmt = _("The number of VFC slots on the target system "
                "(%(rebuild_slots)d) does not match the number of slots on "
                "the client system (%(original_slots)d).  Unable to rebuild "
                "this virtual machine on this system.")


class InvalidVirtualNetworkDeviceType(AbstractMsgFmtError):
    msg_fmt = _("To register the slot information of the network device a "
                "CNA or VNIC adapter is needed. Instead the following "
                "was given: %(wrapper)s.")


class NotEnoughActiveVioses(AbstractMsgFmtError):
    msg_fmt = _("There are not enough active Virtual I/O Servers available. "
                "Expected %(exp)d; found %(act)d.")


class ViosNotAvailable(AbstractMsgFmtError):
    msg_fmt = _("No Virtual I/O Servers are available.  Attempted to wait for "
                "a VIOS to become active for %(wait_time)d seconds.  Please "
                "check the RMC connectivity between the PowerVM NovaLink and "
                "the Virtual I/O Servers.")


class NoRunningSharedSriovAdapters(AbstractMsgFmtError):
    # sriov_loc_mode_state should be a string comprising one SRIOV adapter per
    # line, each line comprising the physical location code, the mode, and the
    # state, separated by ' | '.
    msg_fmt = _("Could not find any SR-IOV adapters in Sriov mode and Running "
                "state.\nLocation | Mode | State\n%(sriov_loc_mode_state)s")


class InsufficientSRIOVCapacity(AbstractMsgFmtError):
    msg_fmt = _("Unable to fulfill redundancy requirement of %(red)d.  Found "
                "%(found_vfs)d viable backing device(s).")


class SystemNotVNICCapable(AbstractMsgFmtError):
    msg_fmt = _("The Managed System is not vNIC capable.")


class NoVNICCapableVIOSes(AbstractMsgFmtError):
    msg_fmt = _("There are no active vNIC-capable VIOSes.")


class VNICFailoverNotSupportedSys(AbstractMsgFmtError):
    msg_fmt = _("A redundancy of %(red)d was specified, but the Managed "
                "System is not vNIC failover capable.")


class VNICFailoverNotSupportedVIOS(AbstractMsgFmtError):
    msg_fmt = _("A redundancy of %(red)d was specified, but there are no "
                "active vNIC failover-capable VIOSes.")


class NoMediaRepoVolumeGroupFound(AbstractMsgFmtError):
    msg_fmt = _("Unable to locate the volume group %(vol_grp)s to store the "
                "virtual optical media within.  Unable to create the "
                "media repository.")


class CantUpdatePPortsInUse(AbstractMsgFmtError):
    msg_fmt = _("The ManagedSystem update was not attempted because changes "
                "were requested to one or more SR-IOV physical ports which "
                "are in use by vNICs.\n%(warnings)s")


class VNCBasedTerminalFailedToOpen(AbstractMsgFmtError):
    msg_fmt = _("Unable to create VNC based virtual terminal: %(err)s")


class CacheNotSupportedException(AbstractMsgFmtError):
    msg_fmt = _("The Adapter cache is not supported.")


class InvalidEnumValue(AbstractMsgFmtError):
    msg_fmt = _("Invalid value '%(value)s' for '%(enum)s'.  Valid values are: "
                "%(valid_values)s")


class VIOSNotFound(AbstractMsgFmtError):
    msg_fmt = _("No VIOS found with name %(vios_name)s.")


class VGNotFound(AbstractMsgFmtError):
    msg_fmt = _("No volume group found with name %(vg_name)s.")



# Copyright 2015, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for pypowervm.tasks.partition."""

import mock
import testtools

import pypowervm.const as c
import pypowervm.exceptions as ex
import pypowervm.tasks.partition as tpar
import pypowervm.tests.tasks.util as tju
import pypowervm.tests.test_fixtures as fx
import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.logical_partition as lpar
import pypowervm.wrappers.virtual_io_server as vios


LPAR_FEED_WITH_MGMT = 'lpar.txt'
VIO_FEED_WITH_MGMT = 'fake_vios_feed.txt'
LPAR_FEED_NO_MGMT = 'lpar_ibmi.txt'
VIO_FEED_NO_MGMT = 'fake_vios_feed2.txt'


def mock_vios(name, state, rmc_state, is_mgmt=False, uptime=3601):
    ret = mock.Mock()
    ret.configure_mock(name=name, state=state, rmc_state=rmc_state,
                       is_mgmt_partition=is_mgmt, uptime=uptime)
    return ret


class TestPartition(testtools.TestCase):

    def setUp(self):
        super(TestPartition, self).setUp()
        self.adpt = self.useFixture(
            fx.AdapterFx(traits=fx.RemotePVMTraits)).adpt
        self.mgmt_vio = tju.load_file(VIO_FEED_WITH_MGMT, self.adpt)
        self.mgmt_lpar = tju.load_file(LPAR_FEED_WITH_MGMT, self.adpt)
        self.nomgmt_vio = tju.load_file(VIO_FEED_NO_MGMT, self.adpt)
        self.nomgmt_lpar = tju.load_file(LPAR_FEED_NO_MGMT, self.adpt)

    def test_get_mgmt_lpar(self):
        "Happy path where the LPAR is the mgmt VM is a LPAR."
        self.adpt.read.side_effect = [self.nomgmt_vio, self.mgmt_lpar]

        mgmt_w = tpar.get_mgmt_partition(self.adpt)
        self.assertTrue(mgmt_w.is_mgmt_partition)
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', mgmt_w.uuid)
        self.assertIsInstance(mgmt_w, lpar.LPAR)
        self.assertEqual(2, self.adpt.read.call_count)

    def test_get_mgmt_vio(self):
        "Happy path where the LPAR is the mgmt VM is a VIOS."
        self.adpt.read.side_effect = [self.mgmt_vio, self.nomgmt_lpar]

        mgmt_w = tpar.get_mgmt_partition(self.adpt)
        self.assertTrue(mgmt_w.is_mgmt_partition)
        self.assertEqual('7DBBE705-E4C4-4458-8223-3EBE07015CA9', mgmt_w.uuid)
        self.assertIsInstance(mgmt_w, vios.VIOS)
        self.assertEqual(1, self.adpt.read.call_count)

    def test_get_mgmt_none(self):
        """Failure path with no mgmt VMs."""
        self.adpt.read.side_effect = [self.nomgmt_lpar, self.nomgmt_vio]

        self.assertRaises(ex.ManagementPartitionNotFoundException,
                          tpar.get_mgmt_partition, self.adpt)

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.search')
    @mock.patch('pypowervm.util.my_partition_id')
    def test_get_me(self, mock_my_id, mock_lp_search, mock_vio_search):
        """Test get_this_partition()."""
        # Good path - one hit on LPAR
        mock_lp_search.return_value = [lpar.LPAR.wrap(self.mgmt_lpar)[0]]
        mock_vio_search.return_value = []
        mock_my_id.return_value = 9
        my_w = tpar.get_this_partition(self.adpt)
        self.assertEqual(9, my_w.id)
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', my_w.uuid)
        mock_lp_search.assert_called_with(self.adpt, id=9)
        mock_vio_search.assert_called_with(self.adpt, id=9)

        # Good path - one hit on VIOS
        mock_lp_search.reset_mock()
        mock_lp_search.return_value = []
        mock_vio_search.return_value = [vios.VIOS.wrap(self.mgmt_vio)[0]]
        mock_my_id.return_value = 2
        my_w = tpar.get_this_partition(self.adpt)
        self.assertEqual(2, my_w.id)
        self.assertEqual('1300C76F-9814-4A4D-B1F0-5B69352A7DEA', my_w.uuid)
        mock_lp_search.assert_not_called()
        mock_vio_search.assert_called_with(self.adpt, id=2)

        # Bad path - no hits
        mock_lp_search.return_value = []
        mock_vio_search.return_value = []
        self.assertRaises(ex.ThisPartitionNotFoundException,
                          tpar.get_this_partition, self.adpt)

    def test_has_physical_io(self):
        """test partition has physical io."""
        part_w = mock.Mock(io_config=mock.Mock(
            io_slots=[mock.Mock(description='1 Gigabit Ethernet (UTP) 4 '
                                'Port Adapter PCIE  Short')]))
        self.assertTrue(tpar.has_physical_io(part_w))

        part_w = mock.Mock(io_config=mock.Mock(
            io_slots=[mock.Mock(description='test Graphics 3.0 test')]))
        self.assertFalse(tpar.has_physical_io(part_w))

        part_w = mock.Mock(io_config=mock.Mock(io_slots=[]))
        self.assertFalse(tpar.has_physical_io(part_w))


class TestVios(twrap.TestWrapper):
    file = 'fake_vios_feed2.txt'
    wrapper_class_to_test = vios.VIOS

    def setUp(self):
        super(TestVios, self).setUp()
        sleep_p = mock.patch('time.sleep')
        self.mock_sleep = sleep_p.start()
        self.addCleanup(sleep_p.stop)
        vioget_p = mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
        self.mock_vios_get = vioget_p.start()
        self.addCleanup(vioget_p.stop)

    def test_get_active_vioses(self):
        self.mock_vios_get.return_value = self.entries
        vioses = tpar.get_active_vioses(self.adpt)
        self.assertEqual(1, len(vioses))
        self.mock_vios_get.assert_called_once_with(self.adpt, xag=())
        vio = vioses[0]
        self.assertEqual(bp.LPARState.RUNNING, vio.state)
        self.assertEqual(bp.RMCState.ACTIVE, vio.rmc_state)
        self.mock_vios_get.assert_called_once_with(self.adpt, xag=())

        self.mock_vios_get.reset_mock()

        # Test with actual xag.  find_min equal to the number found - works.
        vioses = tpar.get_active_vioses(self.adpt, xag='xaglist', find_min=1)
        self.assertEqual(1, len(vioses))
        vio = vioses[0]
        self.assertEqual(bp.LPARState.RUNNING, vio.state)
        self.assertEqual(bp.RMCState.ACTIVE, vio.rmc_state)
        self.mock_vios_get.assert_called_once_with(self.adpt, xag='xaglist')

        # Violates find_min
        self.assertRaises(ex.NotEnoughActiveVioses, tpar.get_active_vioses,
                          self.adpt, find_min=2)

    def test_get_active_vioses_w_vios_wraps(self):
        mock_vios1 = mock_vios('vios1', 'running', 'active')
        mock_vios2 = mock_vios('vios2', 'running', 'inactive')
        mock_vios3 = mock_vios('mgmt', 'running', 'inactive', is_mgmt=True)
        vios_wraps = [mock_vios1, mock_vios2, mock_vios3]

        vioses = tpar.get_active_vioses(self.adpt, vios_wraps=vios_wraps)
        self.assertEqual(2, len(vioses))
        self.mock_vios_get.assert_not_called()

        # The first should be the mgmt partition
        vio = vioses[0]
        self.assertEqual(bp.LPARState.RUNNING, vio.state)
        self.assertEqual(bp.RMCState.INACTIVE, vio.rmc_state)

        # The second should be the active one
        vio = vioses[1]
        self.assertEqual(bp.LPARState.RUNNING, vio.state)
        self.assertEqual(bp.RMCState.ACTIVE, vio.rmc_state)
        self.mock_vios_get.assert_not_called()

    def test_get_physical_wwpns(self):
        self.mock_vios_get.return_value = self.entries
        expected = {'21000024FF649104'}
        result = set(tpar.get_physical_wwpns(self.adpt))
        self.assertSetEqual(expected, result)

    @mock.patch('pypowervm.tasks.partition.get_active_vioses')
    @mock.patch('pypowervm.utils.transaction.FeedTask')
    def test_build_active_vio_feed_task(self, mock_feed_task,
                                        mock_get_active_vioses):
        mock_get_active_vioses.return_value = ['vios1', 'vios2']
        mock_feed_task.return_value = 'mock_feed'
        self.assertEqual('mock_feed', tpar.build_active_vio_feed_task('adpt'))
        mock_get_active_vioses.assert_called_once_with(
            'adpt', xag=(c.XAG.VIO_STOR, c.XAG.VIO_SMAP, c.XAG.VIO_FMAP),
            find_min=1)

    @mock.patch('pypowervm.tasks.partition.get_active_vioses')
    def test_build_tx_feed_task_w_empty_feed(self, mock_get_active_vioses):
        mock_get_active_vioses.return_value = []
        self.assertRaises(
            ex.FeedTaskEmptyFeed, tpar.build_active_vio_feed_task,
            mock.MagicMock())

    def _mk_mock_vioses(self):
        # No
        mock_vios1 = mock_vios('vios1', bp.LPARState.NOT_ACTIVATED,
                               bp.RMCState.INACTIVE)
        # No
        mock_vios2 = mock_vios('vios2', bp.LPARState.RUNNING,
                               bp.RMCState.BUSY)
        # Yes
        mock_vios3 = mock_vios('vios3', bp.LPARState.RUNNING,
                               bp.RMCState.UNKNOWN)
        # No
        mock_vios4 = mock_vios('vios4', bp.LPARState.UNKNOWN,
                               bp.RMCState.ACTIVE)
        # No
        mock_vios5 = mock_vios('vios5', bp.LPARState.RUNNING,
                               bp.RMCState.ACTIVE)
        # Yes
        mock_vios6 = mock_vios('vios6', bp.LPARState.RUNNING,
                               bp.RMCState.INACTIVE)

        # No
        mock_vios7 = mock_vios('vios7', bp.LPARState.RUNNING,
                               bp.RMCState.INACTIVE, is_mgmt=True)

        return [mock_vios1, mock_vios2, mock_vios3, mock_vios4, mock_vios5,
                mock_vios6, mock_vios7]

    @mock.patch('pypowervm.tasks.partition.LOG.warning')
    def test_timeout_short(self, mock_warn):
        """Short timeout because relevant VIOSes have been up a while."""

        self.mock_vios_get.return_value = self._mk_mock_vioses()

        tpar.validate_vios_ready('adap')
        # We slept 120s, (24 x 5s) because all VIOSes have been up >1h
        self.assertEqual(24, self.mock_sleep.call_count)
        self.mock_sleep.assert_called_with(5)
        # We wound up with rmc_down_vioses
        mock_warn.assert_called_once_with(mock.ANY, {'time': 120,
                                                     'vioses': 'vios3, vios6'})
        # We didn't raise - because some VIOSes were okay.

    @mock.patch('pypowervm.tasks.partition.LOG.warning')
    def test_rmc_down_vioses(self, mock_warn):
        """Time out waiting for up/inactive partitions, but succeed."""

        vioses = self._mk_mock_vioses()
        # This one booted "recently"
        vioses[5].uptime = 3559
        self.mock_vios_get.return_value = vioses

        tpar.validate_vios_ready('adap')
        # We slept 600s, (120 x 5s) because one VIOS booted "recently"
        self.assertEqual(120, self.mock_sleep.call_count)
        self.mock_sleep.assert_called_with(5)
        # We wound up with rmc_down_vioses
        mock_warn.assert_called_once_with(mock.ANY, {'time': 600,
                                                     'vioses': 'vios3, vios6'})
        # We didn't raise - because some VIOSes were okay.

    @mock.patch('pypowervm.tasks.partition.LOG.warning')
    def test_no_vioses(self, mock_warn):
        """In the (highly unusual) case of no VIOSes, no warning, but raise."""
        self.mock_vios_get.return_value = []
        self.assertRaises(ex.ViosNotAvailable, tpar.validate_vios_ready, 'adp')
        mock_warn.assert_not_called()

    @mock.patch('pypowervm.tasks.partition.LOG.warning')
    def test_exception_and_good_path(self, mock_warn):
        """VIOS.get raises, then succeeds with some halfsies, then succeeds."""
        vios1_good = mock_vios('vios1', bp.LPARState.RUNNING,
                               bp.RMCState.BUSY)
        vios2_bad = mock_vios('vios2', bp.LPARState.RUNNING,
                              bp.RMCState.UNKNOWN)
        vios2_good = mock_vios('vios2', bp.LPARState.RUNNING,
                               bp.RMCState.ACTIVE)
        self.mock_vios_get.side_effect = (ValueError('foo'),
                                          [vios1_good, vios2_bad],
                                          [vios1_good, vios2_good])
        tpar.validate_vios_ready('adap')
        self.assertEqual(3, self.mock_vios_get.call_count)
        self.assertEqual(2, self.mock_sleep.call_count)
        mock_warn.assert_called_once_with(mock.ANY)

    @mock.patch('pypowervm.tasks.partition.get_mgmt_partition')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.get')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
    def test_get_partitions(self, mock_vio_get, mock_lpar_get, mock_mgmt_get):
        adpt = mock.Mock()

        # Test with the MGMT as a VIOS
        mgmt = mock.Mock(uuid='1')
        vioses = [mock.Mock(uuid='2'), mgmt]
        lpars = [mock.Mock(uuid='3'), mock.Mock(uuid='4')]

        mock_mgmt_get.return_value = mgmt
        mock_vio_get.return_value = vioses
        mock_lpar_get.return_value = lpars

        # Basic case
        self.assertEqual(vioses + lpars, tpar.get_partitions(adpt))

        # Different permutations
        self.assertEqual(lpars + [mgmt], tpar.get_partitions(
            adpt, vioses=False, mgmt=True))
        self.assertEqual(vioses, tpar.get_partitions(
            adpt, lpars=False, mgmt=True))

        # Now test with the MGMT as a LPAR
        vioses = [mock.Mock(uuid='2')]
        lpars = [mock.Mock(uuid='3'), mock.Mock(uuid='4'), mgmt]

        mock_vio_get.return_value = vioses
        mock_lpar_get.return_value = lpars

        # Basic case
        self.assertEqual(vioses + lpars, tpar.get_partitions(adpt))

        # Different permutations
        self.assertEqual(lpars, tpar.get_partitions(
            adpt, vioses=False, mgmt=True))
        self.assertEqual(vioses + [mgmt], tpar.get_partitions(
            adpt, lpars=False, mgmt=True))



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

import pypowervm.adapter as adp
from pypowervm.tests.test_utils import pvmhttp


def refresh_response(file_to_refresh):
    """Reload the file and redo the query."""
    print("Loading original file: ", file_to_refresh)
    new_http = pvmhttp.load_pvm_resp(file_to_refresh)
    if new_http is None or new_http.refresh() is False:
            print("Unable to refresh ", file_to_refresh)
            return 1

    print("Saving refreshed file: ", file_to_refresh)
    new_http.save(file_to_refresh)
    return 0


def usage():
    print("create_httpresp -path path -output out_file [-comment comment]")
    print("    Note: out_file can be a full path or a file in the same "
          "location as create_httpresp.py")

    print('Ex: create_httpresp -path ManagedSystem/<uuid>/LogicalPartition '
          ' -output fakelpar.txt -comment "Created by jsmith"')
    print()
    print('create_httpresp -refresh response_file')
    print('Update a previously created response file by '
          'redoing the same request')

    exit(-1)


def main(argv):

    new_response = pvmhttp.PVMResp()
    output_file = None
    file_to_refresh = None

    aindex = 0
    while aindex < len(argv):
        if argv[aindex] == '-path':
            aindex += 1
            new_response.path = argv[aindex]
        elif argv[aindex] == '-comment':
            aindex += 1
            new_response.comment = argv[aindex]
        elif argv[aindex] == '-output':
            aindex += 1
            output_file = argv[aindex]
        elif argv[aindex] == '-refresh':
            aindex += 1
            file_to_refresh = argv[aindex]
        else:
            print("Unknown argument ", argv[aindex])
            usage()

        aindex += 1

    if file_to_refresh:
        rc = refresh_response(file_to_refresh)
        exit(rc)

    if new_response.path is None or output_file is None:
        usage()

    print("Connecting.")
    adap = adp.Adapter()
    print("Reading path:  ", new_response.path)
    new_response.response = adap.read(new_response.path)

    print("Received ", new_response.response)

    orig_file_name = output_file

    dirname = os.path.dirname(output_file)
    if dirname is None or dirname == '':
        dirname = os.path.dirname(__file__)
        output_file = os.path.join(dirname, output_file)

    new_response.save(output_file)

    print("Response has been saved in ", output_file)
    print("Use the pvmhttp.load_pvm_resp('%s') method "
          "to load it in your testcase " % orig_file_name)

    print("You can have the %s file rebuilt by running: "
          "create_httpresp -refresh %s" %
          (orig_file_name, orig_file_name))


if __name__ == '__main__':
    main(sys.argv[1:])



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Wrappers to parse the output of the PCM JSON data from PHYP."""

import abc
import json

import six

from pypowervm.wrappers import pcm


class PhypInfo(object):
    """Represents a monitor sample from the PHYP monitor.

    The PHYP PCM monitor JSON data can be parsed by this.  The base structure
    is:
    - PhypInfo
      - Info
      - PhypSample
        - PhypSystemFirmware
        - PhypSystemProcessor
        - PhypSystemMemory
        - PhypSharedProcPool
        - PhypVMSample (list - client virtual machines)
          - PhypLparProc
          - PhypLparMemory
          - PhypNetwork
           - PhypVEA
           - PhypSriovLparPort
          - PhypStorage
           - PhypStorageVAdpt
           - PhypVirtualFCAdpt
        - PhypViosSample (list - Virtual I/O Servers)
          - PhypLparProc
    """

    def __init__(self, raw_json):
        data = json.loads(raw_json)
        systemUtil = data.get('systemUtil')
        self.info = pcm.Info(systemUtil.get('utilInfo'))
        self.sample = PhypSample(systemUtil.get('utilSample'))


class PhypSample(object):
    """A Power Hypervisor Sample."""

    def __init__(self, util_sample):
        self.time_stamp = util_sample.get('timeStamp')
        self.status = util_sample.get('status')
        self.time_based_cycles = util_sample.get('timeBasedCycles')

        # Complex objects
        sys_f = util_sample.get('systemFirmware')
        self.system_firmware = (None if sys_f is None
                                else PhypSystemFirmware(sys_f))

        proc = util_sample.get('processor')
        self.processor = None if proc is None else PhypSystemProcessor(proc)

        mem = util_sample.get('memory')
        self.memory = None if mem is None else PhypSystemMemory(mem)

        spp_list = util_sample.get('sharedProcessorPool')
        self.shared_proc_pools = [PhypSharedProcPool(x) for x in spp_list]

        # List of LPARs
        lpars = util_sample.get('lparsUtil')
        self.lpars = [PhypVMSample(x) for x in lpars]

        # List of Virtual I/O Servers
        vioses = util_sample.get('viosUtil')
        self.vioses = [PhypViosSample(x) for x in vioses]


class PhypSystemFirmware(object):
    """Firmware information from PHYP."""

    def __init__(self, system_firmware):
        self.utilized_proc_cycles = system_firmware.get('utilizedProcCycles')
        self.assigned_mem = system_firmware.get('assignedMem')


class PhypSharedProcPool(object):
    """Information of the Shared Processor Pool."""

    def __init__(self, spp):
        self.id = spp.get('id')
        self.name = spp.get('name')
        self.assigned_proc_cycles = spp.get('assignedProcCycles')
        self.utilized_pool_cycles = spp.get('utilizedPoolCycles')
        self.max_proc_units = spp.get('maxProcUnits')
        self.borrowed_pool_proc_units = spp.get('borrowedPoolProcUnits')


class PhypSystemProcessor(object):
    """Processor information about the entire system from PHYP."""

    def __init__(self, processor):
        self.total_proc_units = processor.get('totalProcUnits')
        self.configurable_proc_units = processor.get('configurableProcUnits')
        self.available_proc_units = processor.get('availableProcUnits')
        self.proc_cycles_per_sec = processor.get('procCyclesPerSecond')


class PhypSystemMemory(object):
    """System wide Memory information from PHYP."""

    def __init__(self, mem):
        self.total_mem = mem.get('totalMem')
        self.available_mem = mem.get('availableMem')
        self.configurable_mem = mem.get('configurableMem')


@six.add_metaclass(abc.ABCMeta)
class PhypLparSample(object):
    """A LPAR sample presented by PHYP.  Generic for VIOS & VM."""

    def __init__(self, lpar):
        self.id = lpar.get('id')
        self.uuid = lpar.get('uuid')
        self.name = lpar.get('name')
        self.state = lpar.get('state')
        self.affinity_score = lpar.get('affinityScore')

        # Complex types
        proc = lpar.get('processor')
        self.processor = None if proc is None else PhypLparProc(proc)


class PhypViosSample(PhypLparSample):
    """A VIOS sample presented by the PHYP metrics."""

    def __init__(self, lpar):
        super(PhypViosSample, self).__init__(lpar)


class PhypVMSample(PhypLparSample):
    """A Virtual Machine (non VIOS) presented by the PHYP metrics."""

    def __init__(self, lpar):
        super(PhypVMSample, self).__init__(lpar)
        self.type = lpar.get('type')

        # Complex Types
        mem = lpar.get('memory')
        self.memory = None if mem is None else PhypLparMemory(mem)

        net = lpar.get('network')
        self.network = None if net is None else PhypNetwork(net)

        storage = lpar.get('storage')
        self.storage = None if storage is None else PhypStorage(storage)


class PhypLparMemory(object):
    """A sample of a Client Virtual Machines's memory presented by PHYP.

    Part of the PhypLparSample.
    """

    def __init__(self, mem):
        self.logical_mem = mem.get('logicalMem')
        self.backed_physical_mem = mem.get('backedPhysicalMem')


class PhypLparProc(object):
    """A sample of the LPARs processor presented by PHYP.

    Part of the PhypLparSample
    """

    def __init__(self, proc):
        self.pool_id = proc.get('poolId')
        self.mode = proc.get('mode')
        self.virt_procs = proc.get('maxVirtualProcessors')
        self.proc_units = proc.get('maxProcUnits')
        self.weight = proc.get('weight')
        self.entitled_proc_cycles = proc.get('entitledProcCycles')
        self.util_cap_proc_cycles = proc.get('utilizedCappedProcCycles')
        self.util_uncap_proc_cycles = proc.get('utilizedUnCappedProcCycles')
        self.idle_proc_cycles = proc.get('idleProcCycles')
        self.donated_proc_cycles = proc.get('donatedProcCycles')
        self.time_wait_dispatch = proc.get('timeSpentWaitingForDispatch')
        self.total_instructions = proc.get('totalInstructions')
        self.total_inst_exec_time = proc.get('totalInstructionsExecutionTime')


class PhypNetwork(object):
    """A sample of the LPARs network information.

    Part of the PhypLparSample
    """

    def __init__(self, network):
        veas = network.get('virtualEthernetAdapters')
        self.veas = [] if veas is None else [PhypVEA(x) for x in veas]

        sriov_ports = network.get('sriovLogicalPorts')
        self.sriov_ports = ([] if sriov_ports is None
                            else [PhypSriovLparPort(x) for x in sriov_ports])


class PhypVEA(object):
    """The Virtual Ethernet Adapters (aka. CNA's) data."""

    def __init__(self, vea):
        self.vlan_id = vea.get('vlanId')
        self.vswitch_id = vea.get('vswitchId')
        self.physical_location = vea.get('physicalLocation')
        self.is_pvid = vea.get('isPortVLANID')
        self.received_packets = vea.get('receivedPackets')
        self.sent_packets = vea.get('sentPackets')
        self.dropped_packets = vea.get('droppedPackets')
        self.sent_bytes = vea.get('sentBytes')
        self.received_bytes = vea.get('receivedBytes')
        self.received_physical_packets = vea.get('receivedPhysicalPackets')
        self.sent_physical_packets = vea.get('sentPhysicalPackets')
        self.dropped_physical_packets = vea.get('droppedPhysicalPackets')
        self.sent_physical_bytes = vea.get('sentPhysicalBytes')
        self.received_physical_bytes = vea.get('receivedPhysicalBytes')


class PhypSriovLparPort(object):
    """A metric for the SR-IOV Logical Ports."""

    def __init__(self, sriov_p):
        self.drc_index = sriov_p.get('drcIndex')
        self.phys_drc_index = sriov_p.get('physicalDrcIndex')
        self.phys_port_id = sriov_p.get('physicalPortId')
        self.physical_location = sriov_p.get('physicalLocation')
        self.received_packets = sriov_p.get('receivedPackets')
        self.sent_packets = sriov_p.get('sentPackets')
        self.dropped_sent_packets = sriov_p.get('droppedSentPackets')
        self.dropped_received_packets = sriov_p.get('droppedReceivedPackets')
        self.sent_bytes = sriov_p.get('sentBytes')
        self.recevied_bytes = sriov_p.get('receivedBytes')
        self.error_in = sriov_p.get('errorIn')
        self.error_out = sriov_p.get('errorOut')


class PhypStorage(object):
    """A sample of the LPARs storage information.

    Part of the PhypLparSample
    """

    def __init__(self, stor):
        v_adpts = stor.get('genericVirtualAdapters')
        self.v_stor_adpts = ([] if v_adpts is None
                             else [PhypStorageVAdpt(x) for x in v_adpts])

        v_fcs = stor.get('virtualFiberChannelAdapters')
        self.v_fc_adpts = ([] if v_fcs is None
                           else [PhypVirtualFCAdpt(x) for x in v_fcs])


class PhypStorageVAdpt(object):
    """An indicator to the Client VM Storage to the VIOS storage elem."""

    def __init__(self, stor):
        self.physical_location = stor.get('physicalLocation')
        self.vios_id = stor.get('viosId')
        self.vios_slot = stor.get('viosAdapterSlotId')


class PhypVirtualFCAdpt(object):
    """An indicator to identify the Client VFC Adpt with the VIOS storage."""

    def __init__(self, vfc):
        self.vios_id = vfc.get('viosId')
        # The PCM metrics will have wwpnPair as key name in older versions
        # and wwpnpair as key name in newer versions.
        self.wwpn_pair = vfc.get('wwpnpair', vfc.get('wwpnPair', []))
        self.physical_location = vfc.get('physicalLocation')



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the raw VIOS long term metrics."""

import testtools

from pypowervm.tests.test_utils import pvmhttp
from pypowervm.wrappers.pcm import vios as pcm_vios

VIOS_DATA = 'vios_pcm_data.txt'


class TestViosLTM(testtools.TestCase):

    def setUp(self):
        super(TestViosLTM, self).setUp()

        self.raw_json = pvmhttp.PVMFile(VIOS_DATA).body

    def test_parse(self):
        info = pcm_vios.ViosInfo(self.raw_json)
        self.assertIsNotNone(info)

        # Validate the info
        self.assertEqual('1.0.0', info.info.version)
        self.assertEqual('Raw', info.info.metric_type)
        self.assertEqual('LTM', info.info.monitoring_type)
        self.assertEqual('8247-22L*2125D4A', info.info.mtms)

        # Validate some samples
        sample = info.sample
        self.assertEqual(u'2015-05-27T00:22:00+0000', sample.time_stamp)
        self.assertEqual(1, sample.id)
        self.assertEqual('IOServer - SN2125D4A', sample.name)

        # Validate Memory
        self.assertEqual(1715, sample.mem.utilized_mem)

        # Validate the Network
        self.assertEqual(6, len(sample.network.adpts))
        self.assertEqual(1, len(sample.network.seas))

        phys_dev = sample.network.adpts[1]
        self.assertEqual('ent0', phys_dev.name)
        self.assertEqual('physical', phys_dev.type)
        self.assertEqual('U78CB.001.WZS007Y-P1-C10-T1',
                         phys_dev.physical_location)
        self.assertEqual(1703083, phys_dev.received_packets)
        self.assertEqual(65801, phys_dev.sent_packets)
        self.assertEqual(0, phys_dev.dropped_packets)
        self.assertEqual(187004823, phys_dev.received_bytes)
        self.assertEqual(71198950, phys_dev.sent_bytes)

        # SEA validation
        sea = sample.network.seas[0]
        self.assertEqual('ent6', sea.name)
        self.assertEqual('sea', sea.type)
        self.assertEqual('U8247.22L.2125D4A-V1-C12-T1', sea.physical_location)
        self.assertEqual(0, sea.received_packets)
        self.assertEqual(0, sea.sent_packets)
        self.assertEqual(0, sea.dropped_packets)
        self.assertEqual(0, sea.received_bytes)
        self.assertEqual(0, sea.sent_bytes)
        self.assertEqual(['ent3', 'ent5'], sea.bridged_adpts)

        # Storage - FC Validation
        fc = sample.storage.fc_adpts[0]
        self.assertEqual('fcs0', fc.name)
        self.assertEqual('21000024ff649104', fc.wwpn)
        self.assertEqual('U78CB.001.WZS007Y-P1-C3-T1', fc.physical_location)
        self.assertEqual(0, fc.num_reads)
        self.assertEqual(0, fc.num_writes)
        self.assertEqual(0, fc.read_bytes)
        self.assertEqual(0, fc.write_bytes)
        self.assertEqual(8, fc.running_speed)

        # VFC Validation
        vfc = sample.storage.fc_adpts[1].ports[0]
        self.assertEqual("vfc1", vfc.name)
        self.assertEqual("21000024ff649159", vfc.wwpn)
        self.assertEqual(1234, vfc.num_reads)
        self.assertEqual(1235, vfc.num_writes)
        self.assertEqual(184184, vfc.read_bytes)
        self.assertEqual(138523, vfc.write_bytes)
        self.assertEqual(8, vfc.running_speed)
        self.assertEqual("U78CB.001.WZS007Y-P1-C3-T2000",
                         vfc.physical_location)

        # Physical Adpt Validation
        padpt = sample.storage.phys_adpts[0]
        self.assertEqual('sissas0', padpt.name)
        self.assertEqual('U78CB.001.WZS007Y-P1-C14-T1',
                         padpt.physical_location)
        self.assertEqual(1089692, padpt.num_reads)
        self.assertEqual(1288936, padpt.num_writes)
        self.assertEqual(557922304, padpt.read_bytes)
        self.assertEqual(659935232, padpt.write_bytes)
        self.assertEqual('sas', padpt.type)

        # Storage Virtual Adapter Validation
        vadpt = sample.storage.virt_adpts[0]
        self.assertEqual('vhost5', vadpt.name)
        self.assertEqual('U8247.22L.2125D4A-V1-C7', vadpt.physical_location)
        self.assertEqual(0, vadpt.num_reads)
        self.assertEqual(1, vadpt.num_writes)
        self.assertEqual(0, vadpt.read_bytes)
        self.assertEqual(512, vadpt.write_bytes)
        self.assertEqual('virtual', vadpt.type)

        # SSP Validation
        ssp = sample.storage.ssps[0]
        self.assertEqual('ssp1', ssp.name)
        self.assertEqual(["sissas0"], ssp.pool_disks)
        self.assertEqual(12346, ssp.num_reads)
        self.assertEqual(17542, ssp.num_writes)
        self.assertEqual(18352435, ssp.total_space)
        self.assertEqual(123452, ssp.used_space)
        self.assertEqual(123825, ssp.read_bytes)
        self.assertEqual(375322, ssp.write_bytes)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import logging as base_logging
import mock
import testtools

import pypowervm.adapter as adp
import pypowervm.const as c
import pypowervm.exceptions as pvmex
import pypowervm.helpers.log_helper as log_hlp
import pypowervm.tests.test_fixtures as fx

# Testing by hand it's useful to enable the next line instead of the following
# logging.basicConfig(level=logging.INFO)
base_logging.basicConfig()


class TestLogHelper(testtools.TestCase):

    def setUp(self):
        super(TestLogHelper, self).setUp()
        self.sess = self.useFixture(fx.SessionFx()).sess

    @mock.patch('pypowervm.helpers.log_helper.LOG')
    def test_log_helper(self, mock_log):

        helpers = log_hlp.log_helper
        response = adp.Response('GET', '/some/path', 200, 'OK', ['headers'])
        self.sess.request.return_value = response
        adpt = adp.Adapter(self.sess, helpers=helpers)

        # Test that we get the response we expect passed back unharmed
        self.assertEqual(response,
                         adpt._request('method', 'path', body='the body'))

        # Should be 1 req/resp in the log now, which would be 4 info messages
        mock_log.reset_mock()
        log_hlp._write_thread_log()
        self.assertEqual(mock_log.info.call_count, 4)

        # Should be empty now
        mock_log.reset_mock()
        log_hlp._write_thread_log()
        self.assertEqual(mock_log.info.call_count, 0)

        # Test that we limit the number of entries
        mock_log.reset_mock()
        for x in range(0, 30):
            adpt._request('method1', 'path', body='the body %d' % x)
        log_hlp._write_thread_log()
        # Each req/resp pair is 2 log entries but headers and body
        # are logged separately, so with maxlogs=3, it's 3 * 2 * 2.
        self.assertEqual(mock_log.info.call_count, (3 * 2 * 2))

        mock_log.reset_mock()
        # Add a few records
        adpt._request('method1', 'path', body='the body')
        # Ensure a 412 (special case) doesn't dump, but does raise
        self.sess.request.side_effect = pvmex.HttpError(mock.Mock(
            status=c.HTTPStatus.ETAG_MISMATCH))
        self.assertRaises(
            pvmex.HttpError, adpt._request, 'method2', 'path', body='the body')
        self.assertEqual(0, mock_log.info.call_count)
        # Ensure a non-412 exception dumps the logs and is then raised
        self.sess.request.side_effect = pvmex.HttpError(mock.Mock(
            status=c.HTTPStatus.INTERNAL_ERROR))
        mock_log.reset_mock()
        self.assertRaises(
            pvmex.Error, adpt._request, 'method', 'path', body='the body')
        # Should be 10 entries. 4 * 2 req/resp, 2 for this req.
        self.assertEqual(mock_log.info.call_count, 10)

        # Ensure the log storage is initialized correctly, and we can change
        # the default value
        hlp_size = functools.partial(log_hlp.log_helper, max_logs=12)
        adpt1 = adp.Adapter(self.sess, helpers=hlp_size)
        self.sess.request.side_effect = None
        with mock.patch('pypowervm.helpers.log_helper.'
                        '_init_thread_stg') as mock_init:
            adpt1._request('method1', 'path', body='the body')
            # Should be called with 24 since 12 * 2 entries.
            self.assertEqual(mock_init.call_args_list,
                             [mock.call(max_entries=24)])



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import six

datapath = os.path.join(os.path.dirname(__file__), "data")


def file2b(basename):
    """Reads a file into a byte string.

    :param basename: The base name (no path) of the file to consume.  The
        file is expected to reside in the data/ subdirectory of the path
        containing this library.
    :return: Python 2- and 3-compatible byte string of the input file's
        contents, unaltered and unprocessed.
    """
    with open(os.path.join(datapath, basename), "r") as fh:
        return six.b(fh.read())



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

import pypowervm.const as c
from pypowervm.wrappers import job
import pypowervm.wrappers.logical_partition as wlpar

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

_SUFFIX_PARM_MIGRATE = 'Migrate'
_SUFFIX_PARM_MIGRATE_VALIDATE = 'MigrateValidate'
_SUFFIX_PARM_MIGRATE_ABORT = 'MigrateAbort'
_SUFFIX_PARM_MIGRATE_RECOVER = 'MigrateRecover'

TGT_MGD_SYS = 'TargetManagedSystemName'
TGT_RMT_HMC = 'TargetRemoteHMCIPAddress'
TGT_RMT_HMC_USR = 'TargetRemoteHMCUserID'
VFC_MAPPINGS = 'VirtualFCMappings'
VSCSI_MAPPINGS = 'VirtualSCSIMappings'
VLAN_MAPPINGS = 'VlanMappings'
DEST_MSP = 'DestMSPIPaddr'
SRC_MSP = 'SourceMSPIPaddr'
SPP_ID = 'SharedProcPoolID'
OVS_OVERRIDE = 'OVSOverride'
VLAN_BRIDGE_OVERRIDE = 'VLANBridgeOverride'

_OVERRIDE_OK = '2'


def migrate_lpar(
        lpar, tgt_mgd_sys, validate_only=False, tgt_mgmt_svr=None,
        tgt_mgmt_usr=None, virtual_fc_mappings=None,
        virtual_scsi_mappings=None, dest_msp_name=None, source_msp_name=None,
        spp_id=None, timeout=CONF.pypowervm_job_request_timeout * 4,
        sdn_override=False, vlan_check_override=False, vlan_mappings=None):
    """Method to migrate a logical partition.

    :param lpar: The LPAR wrapper of the logical partition to migrate.
    :param tgt_mgd_sys: The name of the managed system to migrate to.
    :param validate_only: Indication of whether to just validate the migration
        or actually perform it.
    :param tgt_mgmt_svr: The ip of the PowerVM management platform managing
        the target host.
    :param tgt_mgmt_usr: The user id to use on the target PowerVM management
        platform.
    :param virtual_fc_mappings: The virtual fiber channel mappings to move
        during the migration.  See information below.
    :param virtual_scsi_mappings: The virtual scsi mappings to move during the
        migration. See information below.
    :param dest_msp_name: The name of the destination VIOS to use for the mover
        partition.
    :param source_msp_name: The name of the source VIOS to use for the mover
        partition.
    :param spp_id: The shared processor pool id to use on the target system.
    :param timeout: maximum number of seconds for job to complete
    :param sdn_override: (Optional, Default: False) If set to True, will allow
                         a migration where the networking is hosted on a non-
                         traditional VIOS partition (ex. the NovaLink)
    :param vlan_check_override: (Optional, Default: False) If set to True, will
                                tell the Virtual I/O Server not to validate
                                that the other VIOS has the VLAN
                                pre-provisioned.
    :param vlan_mappings: The vlan mappings that indicate what the VLAN should
        be on the target system for a given MAC address.  If not provided, the
        original VLANs will be used.  See information below.

    virtual_fc_mappings:

    List of virtual fibre channel adapter mappings, with each mapping having
    the following format:

    virtual-slot-number/vios-lpar-name/vios-lpar-ID
    [/[vios-virtual-slot-number][/[vios-fc-port-name]]]

    The first two '/' characters must be present. The third '/' character is
    optional, but it must be present if vios-virtual-slot-number or
    vios-fc-port-name is specified.  The last '/' character is optional but it
    must be present if vios-fc-port-name is specified.

    Optional values may be omitted. Optional values are vios-lpar-name
    or vios-lpar-ID (one of those values is required, but not both),
    vios-virtual-slot-number, and vios-fc-port-name.

    For example:
    4//1/14/fcs0 specifies a mapping of the virtual fibre channel client
    adapter with slot number 4 to the virtual fibre channel server adapter with
    slot number 14 in the VIOS partition with ID 1 on the destination managed
    system. In addition, the mapping specifies to use physical fibre channel
    port fcs0.

    virtual_scsi_mappings:

    List of virtual SCSI adapter mappings, with each mapping having the
    following format:

    virtual-slot-number/vios-lpar-name/vios-lpar-ID
    [/vios-virtual-slot-number]

    The first two '/' characters must be present.  The last '/' character is
    optional, but it must be present if vios-virtual-slot-number is specified.
    Optional values may be omitted. Optional values are vios-lpar-name or
    vios-lpar-ID (one of those values is required, but not both), and
    vios-virtual-slot-number.

    For example:
    12/vios1//16 specifies a mapping of the virtual SCSI adapter with slot
    number 12 to slot number 16 on the VIOS partition vios1 on the destination
    managed system.

    vlan_mappings:

    List of vlan mappings, with each mapping having the following format:

    MAC/PVID[/VLAN_A VLAN_B]

    The first '/' must be present.  The first field is the MAC address of the
    adapter.  The MAC address must be exactly 12 digits, case insensitive,
    without colons in it.  The second is what the target PVID should be set to
    for that adapter.  The remaining is a list of additional VLANs that could
    be specified for adapters that have additional VLANs.  The list of
    additional VLANs is space delimited.

    For example:
    001122334455/12 specifies a mapping where the adapter with MAC address
    001122334455 should have a PVID of 12 on the target system.
    """

    op = (_SUFFIX_PARM_MIGRATE_VALIDATE
          if validate_only else _SUFFIX_PARM_MIGRATE)
    resp = lpar.adapter.read(wlpar.LPAR.schema_type, lpar.uuid,
                             suffix_type=c.SUFFIX_TYPE_DO,
                             suffix_parm=op)
    job_wrapper = job.Job.wrap(resp.entry)
    job_parms = [job_wrapper.create_job_parameter(TGT_MGD_SYS,
                                                  str(tgt_mgd_sys))]

    # Generic 'raw' format job parameters.
    for kw, val in [(TGT_RMT_HMC, tgt_mgmt_svr),
                    (TGT_RMT_HMC_USR, tgt_mgmt_usr), (DEST_MSP, dest_msp_name),
                    (SRC_MSP, source_msp_name), (SPP_ID, spp_id)]:
        if val:
            job_parms.append(
                job_wrapper.create_job_parameter(kw, str(val)))

    # The SDN / VLAN overrides are...odd.  Instead of passing in a 'True', we
    # must pass in the character of '2' to indicate that it is an override.
    for kw, val in [(OVS_OVERRIDE, sdn_override),
                    (VLAN_BRIDGE_OVERRIDE, vlan_check_override)]:
        if val:
            job_parms.append(job_wrapper.create_job_parameter(kw,
                                                              _OVERRIDE_OK))

    # The mappings are special.  They require a join so that they are comma
    # separated down to the API.
    for kw, val in [(VFC_MAPPINGS, virtual_fc_mappings),
                    (VSCSI_MAPPINGS, virtual_scsi_mappings),
                    (VLAN_MAPPINGS, vlan_mappings)]:
        if val:
            job_parms.append(
                job_wrapper.create_job_parameter(kw, ",".join(val)))

    job_wrapper.run_job(lpar.uuid, job_parms=job_parms, timeout=timeout)


def migrate_recover(lpar, force=False,
                    timeout=CONF.pypowervm_job_request_timeout):

    """Method to recover a failed logical partition migration.

    :param lpar: The LPAR wrapper of the logical partition to recover.
    :param force: Boolean specifying whether to force the migration to recover
        when errors are encountered.
    :param timeout: maximum number of seconds for job to complete
    """
    resp = lpar.adapter.read(wlpar.LPAR.schema_type, lpar.uuid,
                             suffix_type=c.SUFFIX_TYPE_DO,
                             suffix_parm=_SUFFIX_PARM_MIGRATE_RECOVER)
    job_wrapper = job.Job.wrap(resp.entry)
    job_parms = []
    if force:
        job_parms.append(job_wrapper.create_job_parameter('Force', 'true'))

    job_wrapper.run_job(lpar.uuid, job_parms=job_parms, timeout=timeout)


def migrate_abort(lpar, timeout=CONF.pypowervm_job_request_timeout):

    """Method to abort a logical partition migration.

    :param lpar: The LPAR wrapper of the logical partition to abort the
        migration operation.
    :param timeout: maximum number of seconds for job to complete
    """

    resp = lpar.adapter.read(wlpar.LPAR.schema_type, lpar.uuid,
                             suffix_type=c.SUFFIX_TYPE_DO,
                             suffix_parm=_SUFFIX_PARM_MIGRATE_ABORT)
    job_wrapper = job.Job.wrap(resp.entry)
    job_wrapper.run_job(lpar.uuid, job_parms=None, timeout=timeout)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

import pypowervm.tests.test_utils.test_wrapper_abc as twrap
from pypowervm.wrappers import event


class TestEvent(twrap.TestWrapper):
    file = 'event_feed.txt'
    wrapper_class_to_test = event.Event

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.get')
    def test_get(self, mock_ewrap_get):
        event.Event.get('adap', 'appid')
        mock_ewrap_get.assert_called_once_with('adap', xag=[], add_qp=[
            ('QUEUE_CLIENTKEY_METHOD', 'USE_APPLICATIONID'),
            ('QUEUE_APPLICATIONID', 'appid')])

    def test_getters(self):
        ev1, ev2 = self.entries[:2]
        self.assertEqual('510ae1e6-3e86-34c6-bf4c-c638e76a5f68', ev1.uuid)
        self.assertEqual(event.EventType.ADD_URI, ev1.etype)
        self.assertEqual('1473962006548', ev1.eid)
        self.assertEqual('http://localhost:12080/rest/api/uom/ManagedSystem/1c'
                         'ab7366-6b73-342c-9f43-ddfeb9f8edd3/LogicalPartition/'
                         '1E6FC741-6253-4B69-B88B-8A44BED92145', ev1.data)
        self.assertIsNone(ev1.detail)
        self.assertEqual(event.EventType.MODIFY_URI, ev2.etype)
        self.assertEqual('Other', ev2.detail)

    def test_bld(self):
        evt = event.Event.bld('adap', None, None)
        self.assertIsInstance(evt, event.Event)
        self.assertEqual('adap', evt.adapter)
        self.assertIsNone(evt.data)
        self.assertIsNone(evt.detail)
        evt = event.Event.bld('adap2', 'data', 'detail')
        self.assertIsInstance(evt, event.Event)
        self.assertEqual('adap2', evt.adapter)
        self.assertEqual('data', evt.data)
        self.assertEqual('detail', evt.detail)

    def test_str(self):
        self.assertEqual(
            'Event(id=1473962006548, type=ADD_URI, data=http://localhost:12080'
            '/rest/api/uom/ManagedSystem/1cab7366-6b73-342c-9f43-ddfeb9f8edd3/'
            'LogicalPartition/1E6FC741-6253-4B69-B88B-8A44BED92145, detail=Non'
            'e)', str(self.dwrap))
